import datetime
import re

from .utils import ads_reports_parsing
from google.ads.googleads.client import GoogleAdsClient
from google.ads.googleads.v15.services.types.google_ads_service import GoogleAdsRow

import pandas as pd

class AdsWrapper:
    def __init__(self,
                 googleads_client: GoogleAdsClient,
                 customer_id: str,):
        self.client = googleads_client
        self.service = self.client.get_service("GoogleAdsService", version='v15')
        self.customer_id = customer_id

    def get_data(self,
                      resource: str,
                      date: datetime.date | str | tuple[datetime.date | str, datetime.date | str] | None,
                      dimensions: list[str],
                      metrics: list[str],
                      conditions: list | None = None,
                      custom_condition_expr: str | None = None,
                      limit: int | None = None,
                    _return_raw: bool = False,
                 _print_query: bool = False):

        rows = self._make_request(
            resource=resource,
            date=date,
            metrics=metrics,
            dimensions=dimensions,
            conditions=conditions,
            custom_condition_expr=custom_condition_expr,
            limit=limit,
            _return_raw=_return_raw
        )
        if _return_raw:
            return rows
        return ads_reports_parsing.tidy_column_names(
            pd.DataFrame.from_records(rows)
        )


    def _make_request(self,
                     resource: str,
                     date: datetime.date | str | tuple[datetime.date | str, datetime.date | str] | None,
                     dimensions: list[str],
                     metrics: list[str] = None,
                     conditions: list | None = None,
                     custom_condition_expr: str | None = None,
                     limit: int | None = None,
                      _return_raw: bool = False,
                      _print_query: bool = False) -> list[dict | GoogleAdsRow]:

        if metrics is None:
            metrics = []
        metrics = [f"metrics.{m}" if not re.match(r'metrics\.', m) else m for m in metrics]

        if conditions is None:
            conditions = []
        if date is not None:
            _dc = ads_reports_parsing.make_date_condition(dt=date)
            conditions = [_dc] + conditions

        search_request = self.client.get_type("SearchGoogleAdsStreamRequest")
        search_request.customer_id = self.customer_id
        q = ads_reports_parsing.create_google_ads_sql_query(
            resource=resource,
            dimensions=dimensions + metrics,
            conditions=conditions,
            custom_condition_expr=custom_condition_expr,
            limit=limit
        )
        if _print_query:
            print(q)
        search_request.query = re.sub(r'\n', '', q)

        stream = self.service.search_stream(search_request)
        raw_rows = []
        for batch in stream:
            raw_rows.extend([row for row in batch.results])
        if _return_raw:
            return raw_rows
        return ads_reports_parsing.extract_google_ads_rows(
            raw_rows, dimensions=dimensions + metrics
        )










import json
import yaml
import re

from typing import Tuple

from google.oauth2 import service_account  # pip install --upgrade google-auth
from googleapiclient import discovery  # pip install --upgrade google-api-python-client
from google.analytics.data_v1beta import BetaAnalyticsDataClient  # pip install google-analytics-data
from google.ads.googleads.client import GoogleAdsClient

_default_googleads_api_version = 15

def googleads_client_from_key_file(path: str):
    with open(path, 'rb') as _file:
        googleads_yaml_string = _file.read().decode('utf8')
    return googleads_client_from_yaml(googleads_yaml_string=googleads_yaml_string)


def googleads_client_from_yaml(googleads_yaml_string: str) -> Tuple[GoogleAdsClient, str, dict]:
    googleads_yaml_dict = yaml.safe_load(googleads_yaml_string)
    default_customer_id = parse_ads_id(googleads_yaml_dict.get('default_customer_id', ''))
    api_version = parse_api_version(googleads_yaml_dict.get('api_version', _default_googleads_api_version))
    googleads_client = GoogleAdsClient.load_from_string(yaml_str=googleads_yaml_string, version=api_version)
    return googleads_client, default_customer_id, googleads_yaml_dict


def parse_ads_id(customer_id: str | int) -> str:
    if isinstance(customer_id, int):
        customer_id = str(customer_id)
    return re.sub(r'[^\d]', '', customer_id)


def parse_api_version(api_version: str | int) -> str:
    if isinstance(api_version, int):
        _api_version = "v" + str(api_version)
    elif isinstance(api_version, str) and len(api_version) > 0:
        if default_api_version[0] == 'v':
            _api_version = api_version
        else:
            _api_version = "v" + api_version
    else:
        _api_version = "v15"
    return _api_version


def get_analytics_resources(json_api_key: str = None, key_file_path: str = None):
    if key_file_path and not json_api_key:
        with open(key_file_path, 'rb') as _file:
            json_api_key = _file.read().decode('utf8')
    if json_api_key:
        project_credentials = get_analytics_project_credentials(secrets=json.loads(json_api_key))
    else:
        project_credentials = None
    return build_analytics_resources(analytics_project_credentials=project_credentials)


def get_analytics_project_credentials(secrets: dict):
    analytics_project_scopes = ['https://www.googleapis.com/auth/analytics.readonly',
                                'https://www.googleapis.com/auth/webmasters.readonly',
                                'https://www.googleapis.com/auth/webmasters',
                                'https://www.googleapis.com/auth/cloud-platform']

    project_credentials = service_account.Credentials.from_service_account_info(
        info=secrets).with_scopes(scopes=analytics_project_scopes)
    return project_credentials


def build_analytics_resources(analytics_project_credentials: service_account.Credentials = None):
    ga3_resource = discovery.build('analyticsreporting', 'v4', credentials=analytics_project_credentials)
    gsc_resource = discovery.build('searchconsole', 'v1', credentials=analytics_project_credentials)
    ga4_data_client = BetaAnalyticsDataClient(credentials=analytics_project_credentials)
    return ga3_resource, ga4_data_client, gsc_resource



import datetime
import logging
import time

import pandas as pd
import numpy as np
import uuid
import re

from math import ceil
from typing import Any

from google.ads.googleads.client import GoogleAdsClient
# from google.ads.googleads.v10.resources import KeywordPlanGeoTarget
from google.ads.googleads.errors import GoogleAdsException
from google.api_core.exceptions import InternalServerError

from . import LOCATION_ID_DICT

DEFAULT_LOCATIONS = "GBR"  # location defaults GBR
DEFAULT_LANGUAGE_ID = "1000"  # language defaults to "1000" (i.e. English)

DEFAULT_KEYWORD_PLAN_CAMPAIGN_CPC_BID = 1000000
DEFAULT_KEYWORD_PLAN_AD_GROUP_CPC_BID = 250000
API_MULTIPLE_REQUEST_WAIT_TIME = 2

RE_URL = re.compile(r"https?://(www\.)?[\w\-_+]*(\.\w{2,4}){0,2}/")


class ClientWrapper:
    def __init__(self,
                 googleads_client: GoogleAdsClient,
                 customer_id: str,
                 location_codes: list[str] | None = None,
                 language_id: str = None,
                 api_version: str | int = 15
                 ):
        self.client: GoogleAdsClient = googleads_client
        self.customer_id = customer_id

        if isinstance(api_version, int):
            api_version = "v" + str(api_version)
        self.googleads_service = self.client.get_service("GoogleAdsService", version=api_version)

        if location_codes is None:
            location_codes = DEFAULT_LOCATIONS
        if isinstance(location_codes, str):
            location_codes = [location_codes]
        self.location_codes = location_codes

        if language_id is None:
            language_id = DEFAULT_LANGUAGE_ID
        self.language_id = language_id

        self.location_rns = _map_location_to_resource_names(client=self.client,
                                                            location_codes=self.location_codes)
        self.language_rn = _map_language_id_to_resource_name(client=self.client,
                                                             language_id=self.language_id)

    def get_campaigns(self) -> pd.DataFrame:
        query = """
            SELECT
              campaign.id,
              campaign.name
            FROM campaign
            ORDER BY campaign.id
        """

        # Issues a search request using streaming.
        stream = self.googleads_service.search_stream(customer_id=self.customer_id, query=query)

        _out = []
        for batch in stream:
            for row in batch.results:
                _out.append({"id": row.campaign.id,
                             "name": row.campaign.name})
        return pd.DataFrame(_out)

    def get_ad_groups(self, campaign_id=None) -> pd.DataFrame:
        query = """
            SELECT
              campaign.id,
              ad_group.id,
              ad_group.name
            FROM ad_group"""

        if campaign_id:
            query += f" WHERE campaign.id = {campaign_id}"

        search_request = self.client.get_type("SearchGoogleAdsRequest")
        search_request.customer_id = self.customer_id
        search_request.query = query
        search_request.page_size = 1000

        results = self.googleads_service.search(request=search_request)

        _out = []
        for row in results:
            _out.append({"campaign_id": row.campaign.id,
                         "ad_group_name": row.ad_group.name,
                         "ad_group_id": row.ad_group.id})
        return pd.DataFrame(_out)

    def get_keyword_plan_ad_groups(self) -> pd.DataFrame:
        """Return keyword_plan ad groups -- these don't show up as regular ad groups"""

        query = """
            SELECT
              keyword_plan.id,
              keyword_plan.name,
              keyword_plan_ad_group.id,
              keyword_plan_ad_group.name,
              keyword_plan_campaign.id,
              keyword_plan_campaign.name
            FROM keyword_plan_ad_group"""

        search_request = self.client.get_type("SearchGoogleAdsRequest")
        search_request.customer_id = self.customer_id
        search_request.query = query
        search_request.page_size = 1000

        results = self.googleads_service.search(request=search_request)

        _out = []
        for row in results:
            _out.append({"keyword_plan_name": row.keyword_plan.name,
                         "keyword_plan_id": row.keyword_plan.id,
                         "ad_group_name": row.keyword_plan_ad_group.name,
                         "ad_group_id": row.keyword_plan_ad_group.id,
                         "campaign_name": row.keyword_plan_campaign.name,
                         "campaign_id": row.keyword_plan_campaign.id
                         })
        return pd.DataFrame(_out)

    def get_keyword_plan_ad_group_keywords(self,
                                           keyword_plan_resource_name: str = None,
                                           keyword_plan_id: int = None) -> pd.DataFrame:
        query = """
            SELECT
              keyword_plan.id,
              keyword_plan_ad_group_keyword.id,
              keyword_plan_ad_group_keyword.text
            FROM keyword_plan_ad_group_keyword"""
        if keyword_plan_resource_name:
            query += f" WHERE keyword_plan.resource_name = '{keyword_plan_resource_name}'"
        elif keyword_plan_id:
            query += f" WHERE keyword_plan.id = {keyword_plan_id}"

        search_request = self.client.get_type("SearchGoogleAdsRequest")
        search_request.customer_id = self.customer_id
        search_request.query = query
        # search_request.page_size = 1000

        results = self.googleads_service.search(request=search_request)

        _out = []
        for row in results:
            _out.append({"keyword_plan_id": row.keyword_plan.id,
                         "keyword_id": row.keyword_plan_ad_group_keyword.id,
                         "keyword_resource_name": row.keyword_plan_ad_group_keyword.resource_name,
                         "keyword_text": row.keyword_plan_ad_group_keyword.text})
        return pd.DataFrame(_out)

    def get_keyword_plans(self) -> pd.DataFrame:
        """Return keyword_plan ad groups -- these don't show up as regular ad groups"""

        query = """
            SELECT
              keyword_plan.id,
              keyword_plan.name
            FROM keyword_plan"""

        search_request = self.client.get_type("SearchGoogleAdsRequest")
        search_request.customer_id = self.customer_id
        search_request.query = query
        search_request.page_size = 1000

        results = self.googleads_service.search(request=search_request)

        _out = []
        for row in results:
            _out.append({"keyword_plan_id": row.keyword_plan.id,
                         "keyword_plan_name": row.keyword_plan.name,
                         "resource_name": row.keyword_plan.resource_name})

        return pd.DataFrame(_out)

    def _get_keyword_plan_ids(self) -> list[int]:
        """Return keyword_plan ad groups -- these don't show up as regular ad groups"""

        query = """
            SELECT
              keyword_plan.id,
              keyword_plan.name
            FROM keyword_plan"""

        search_request = self.client.get_type("SearchGoogleAdsRequest")
        search_request.customer_id = self.customer_id
        search_request.query = query
        search_request.page_size = 1000

        results = self.googleads_service.search(request=search_request)

        _out = []
        for row in results:
            _out.append(row.keyword_plan.id)

        return _out

    def remove_campaign(self,
                        campaign_id: str = None,
                        campaign_resource_name: str = None):
        # https://developers.google.com/google-ads/api/docs/samples/remove-campaign
        campaign_service = self.client.get_service("CampaignService")
        campaign_operation = self.client.get_type("CampaignOperation")

        if campaign_id is not None:
            resource_name = campaign_service.campaign_path(self.customer_id, campaign_id)
        elif campaign_resource_name is not None:
            resource_name = campaign_resource_name
        else:
            raise ValueError("must supply either campaign_id or campaign_resource_name")

        campaign_operation.remove = resource_name

        campaign_response = campaign_service.mutate_campaigns(
            customer_id=self.customer_id, operations=[campaign_operation]
        )

        print(f"Removed campaign {campaign_response.results[0].resource_name}.")

    def remove_keyword_plan(self,
                            keyword_plan_id: int | str = None,
                            keyword_plan_resource_name: str = None):
        # https://developers.google.com/google-ads/api/docs/samples/remove-campaign
        keyword_plan_service = self.client.get_service("KeywordPlanService")
        keyword_plan_operation = self.client.get_type("KeywordPlanOperation")

        if keyword_plan_id is not None:
            if isinstance(keyword_plan_id, int):
                keyword_plan_id = str(keyword_plan_id)
            resource_name = keyword_plan_service.keyword_plan_path(self.customer_id, keyword_plan_id)
        elif keyword_plan_resource_name is not None:
            resource_name = keyword_plan_resource_name
        else:
            raise ValueError("must supply either campaign_id or campaign_resource_name")

        keyword_plan_operation.remove = resource_name

        keyword_plan_response = keyword_plan_service.mutate_keyword_plans(
            customer_id=self.customer_id, operations=[keyword_plan_operation]
        )

        # print(f"Removed campaign {keyword_plan_response.results[0].resource_name}.")

    def remove_ad_group(self,
                        ad_group_id: str = None,
                        ad_group_resource_name: str = None):
        ad_group_service = self.client.get_service("AdGroupService")
        ad_group_operation = self.client.get_type("AdGroupOperation")

        if ad_group_id is not None:
            resource_name = ad_group_service.ad_group_path(self.customer_id, ad_group_id)
        elif ad_group_resource_name is not None:
            resource_name = ad_group_resource_name
        else:
            raise ValueError("must supply either ad_group_id or ad_group_resource_name")

        ad_group_operation.remove = resource_name

        ad_group_response = ad_group_service.mutate_ad_groups(
            customer_id=self.customer_id, operations=[ad_group_operation]
        )

        # print(f"Removed ad group {ad_group_response.results[0].resource_name}.")

    def _remove_all_keyword_plans(self):
        keyword_plan_ids_list = self._get_keyword_plan_ids()

        for keyword_plan_id in keyword_plan_ids_list:
            self.remove_keyword_plan(keyword_plan_id=keyword_plan_id)


class KeywordPlanService(ClientWrapper):
    def __init__(self,
                 googleads_client: GoogleAdsClient,
                 customer_id: str,
                 location_codes: list[str] = None,
                 language_id: str | None = None
                 ):
        super().__init__(googleads_client=googleads_client,
                         customer_id=customer_id,
                         location_codes=location_codes,
                         language_id=language_id)

        self.keyword_plan_service = self.client.get_service("KeywordPlanService")
        self.keyword_plan_campaign_service = self.client.get_service("KeywordPlanCampaignService")
        self.keyword_plan_ad_group_service = self.client.get_service("KeywordPlanAdGroupService")
        self.keyword_plan_ad_group_keyword_service = self.client.get_service("KeywordPlanAdGroupKeywordService")

        self.default_network = self.client.enums.KeywordPlanNetworkEnum.GOOGLE_SEARCH
        self.default_keyword_plan_campaign_cpc_bid_micros = DEFAULT_KEYWORD_PLAN_CAMPAIGN_CPC_BID
        self.default_keyword_plan_ad_group_cpc_bid_micros = DEFAULT_KEYWORD_PLAN_AD_GROUP_CPC_BID

    def get_keyword_metrics(self,
                            keywords: list[str] | str,
                            location_codes: list[str] = None,
                            language_id: str = None,
                            print_progress: bool = False,
                            calculated_fields: bool = True,
                            concat_locations: bool = True) -> tuple[pd.DataFrame, pd.DataFrame]:

        if isinstance(keywords, str):
            keywords = [keywords]

        if location_codes is None:
            location_codes = self.location_codes

        keyword_list_partition = _partition_list(_list=keywords, n=2_000)
        frames: list[pd.DataFrame] = []
        keyword_plans = []

        for _sublist_index, _keyword_sublist in enumerate(keyword_list_partition):
            if _sublist_index > 0:
                time.sleep(API_MULTIPLE_REQUEST_WAIT_TIME)

            try:
                if concat_locations:
                    historical_metrics_df, _kwps = self.get_historical_metrics_df_concat_locations(
                        keywords=_keyword_sublist,
                        location_codes=location_codes,
                        language_id=language_id,
                        calculated_fields=calculated_fields
                    )
                else:
                    historical_metrics_df, _kwps = self.get_historical_metrics_df(
                        keywords=_keyword_sublist,
                        location_codes=location_codes,
                        language_id=language_id,
                        calculated_fields=calculated_fields
                    )

                if historical_metrics_df is not None:
                    keyword_plans.extend(_kwps)
                    frames.append(historical_metrics_df)
                    for _kwp in _kwps:
                        self.remove_keyword_plan(keyword_plan_resource_name=_kwp)

            except GoogleAdsException:
                print(f"GoogleAdsException encountered after obtaining {len(frames)} frames")
                break

            if print_progress:
                print(f"{len(frames)} - obtained metrics for {len(frames[-1])} keywords")

        if len(frames) > 0:
            _df = pd.concat(frames)
        else:
            return None, None

        metrics_df: pd.DataFrame = _df.drop(columns=['volume_trend_tuples'])

        monthly_volume_df: pd.DataFrame = _df[['country_iso_code',
                                               'date_obtained',
                                               'query',
                                               'keyword',
                                               'status',
                                               'volume_trend_tuples']].explode('volume_trend_tuples')

        monthly_volume_df['volume_trend_tuples'] = monthly_volume_df['volume_trend_tuples'].apply(
            _check_volume_trend_tuples)

        monthly_volume_df['month_name'] = monthly_volume_df['volume_trend_tuples'].apply(lambda t: t[3])
        monthly_volume_df['month_enum'] = monthly_volume_df['volume_trend_tuples'].apply(lambda t: t[2])
        monthly_volume_df['month'] = monthly_volume_df['month_enum'].apply(_get_month_from_enum_value)
        monthly_volume_df['year'] = monthly_volume_df['volume_trend_tuples'].apply(lambda t: t[1])

        monthly_volume_df.dropna(axis='index', subset=['month'], inplace=True)

        monthly_volume_df['record_date'] = monthly_volume_df.apply(
            lambda df: _conv_date(df['year'], df['month'], 15, month_index=0),
            axis=1
        )
        monthly_volume_df['volume'] = monthly_volume_df['volume_trend_tuples'].apply(lambda t: t[0])

        monthly_volume_df.drop(
            columns=['month_enum', 'volume_trend_tuples', 'month', 'year', 'month_name'],
            inplace=True
        )

        metrics_df.drop_duplicates(subset=["country_iso_code", "query", "keyword"], keep="first", inplace=True)
        monthly_volume_df.drop_duplicates(subset=["country_iso_code", "query", "keyword", "record_date"], keep="first", inplace=True)

        metrics_df.reset_index(drop=True, inplace=True)
        monthly_volume_df.reset_index(drop=True, inplace=True)

        _recent_dates = monthly_volume_df[
            ["country_iso_code", 'query', 'keyword', 'record_date']
        ].groupby(by=["country_iso_code", 'query', 'keyword']).max().reset_index()

        _competition_subset = pd.merge(
            left=_recent_dates,
            right=metrics_df[["country_iso_code", 'query', 'keyword', 'competition', 'competition_index']],
            on=["country_iso_code", 'query', 'keyword'],
            how='inner'
        )

        metrics_df = pd.merge(left=metrics_df, right=_recent_dates, on=["country_iso_code", 'query', 'keyword'], how='left')

        monthly_volume_df = pd.merge(
            left=monthly_volume_df,
            right=_competition_subset,
            on=["country_iso_code", 'query', 'keyword', 'record_date'],
            how='left'
        )

        return metrics_df, monthly_volume_df

    def get_forcast_metrics_df(self,
                               keywords: list[str],
                               location_codes: list[str] = None,
                               language_id: str = None) -> tuple[pd.DataFrame, str]:

        keyword_plan, stripped_keyword_dict = self.add_keyword_plan(keywords=keywords,
                                                                    match_type="EXACT",
                                                                    location_codes=location_codes,
                                                                    language_id=language_id)
        metrics = self.generate_forecast_metrics(keyword_plan_resource_name=keyword_plan)
        _df = pd.DataFrame(metrics)

        keywords_df = self.get_keyword_plan_ad_group_keywords(keyword_plan_resource_name=keyword_plan)

        _df["keyword_id"] = _df["keyword_plan_ad_group_keyword"].apply(lambda s: int(s.split('/')[-1]))
        keywords_df["query"] = keywords_df["keyword_text"]
        _df = pd.merge(_df, keywords_df, how="left", on="keyword_id")
        _df = _df[['date_obtained', 'query', 'impressions', 'clicks', 'ctr', 'average_cpc', 'cost_micros']]

        return _df, keyword_plan

    def _get_historical_metrics_df(self,
                                   keywords: list[str],
                                   location_codes: list[str] = None,
                                   language_id: str = None,
                                   calculated_fields: bool = True) -> tuple[pd.DataFrame, str]:
        keyword_plan, stripped_keyword_dict = self.add_keyword_plan(keywords=keywords,
                                                                    match_type="EXACT",
                                                                    location_codes=location_codes,
                                                                    language_id=language_id)
        metrics = self.generate_historical_metrics(keyword_plan_resource_name=keyword_plan,
                                                   stripped_keyword_dict=stripped_keyword_dict)
        # self.remove_keyword_plan(keyword_plan_resource_name=keyword_plan)
        _df = pd.DataFrame(metrics)
        if calculated_fields:
            _df["volume_trend_coef"] = _df["volume_trend"].apply(_three_month_trend_coef)
            _df["volume"] = _df["volume_trend"].apply(_latest_volume)
            _df["volume_last_month"] = _df["volume_trend"].apply(_previous_volume)
            _df["volume_last_year"] = _df["volume_trend"].apply(_last_year_volume)
            _df["volume_3monthavg"] = _df["volume_trend"].apply(_volume_3monthavg)
            _df["volume_6monthavg"] = _df["volume_trend"].apply(_volume_6monthavg)
            _df["volume_12monthavg"] = _df["volume_trend"].apply(_volume_12monthavg)
        return _df, keyword_plan


    def get_historical_metrics_df_concat_locations(self,
                                  keywords: list[str],
                                  location_codes: list[str] = None,
                                  language_id: str = None,
                                  calculated_fields: bool = True,
                                  max_retries: int = 5) -> tuple[pd.DataFrame, list[str]]:

        frames, keyword_plans = [], []

        for _loc_code in location_codes:
            _df, _keyword_plans = self.get_historical_metrics_df(
                keywords=keywords,
                location_codes=[_loc_code],
                language_id=language_id,
                calculated_fields=calculated_fields,
                max_retries=max_retries
            )
            if isinstance(_df, pd.DataFrame):
                frames.append(_df)
            keyword_plans.extend(_keyword_plans)

        if len(frames) > 0:
            dataframe = pd.concat(frames)
        else:
            dataframe = None

        return dataframe, keyword_plans

    def get_historical_metrics_df(self,
                                  keywords: list[str],
                                  location_codes: list[str] = None,
                                  language_id: str = None,
                                  calculated_fields: bool = True,
                                  max_retries: int = 5) -> tuple[pd.DataFrame, list[str]]:

        retries: int = 0
        complete: bool = False
        while retries < max_retries and complete is False:
            try:
                _df, _keyword_plan = self._get_historical_metrics_df(
                    keywords=keywords,
                    location_codes=location_codes,
                    language_id=language_id,
                    calculated_fields=calculated_fields
                )
                complete = True
            except InternalServerError:
                logging.warning("Keyword planner :: InternalServerError :: retrying after 2 seconds")
                time.sleep(2)
                retries += 1

        if complete:
            _plans = [_keyword_plan]
        else:
            _df = None
            _plans = []

        if _df is not None:
            _df.insert(loc=0, column='country_iso_code', value='-'.join(location_codes))

        return _df, _plans

    def generate_forecast_metrics(self,
                                  keyword_plan_resource_name: str):
        response_forcast = self.keyword_plan_service.generate_forecast_metrics(
            keyword_plan=keyword_plan_resource_name,
        )
        _date_today = datetime.date.today()
        metrics = [{"date_obtained": _date_today,
                    "keyword_plan_ad_group_keyword": _m.keyword_plan_ad_group_keyword,
                    "impressions": _m.keyword_forecast.impressions,
                    "clicks": _m.keyword_forecast.clicks,
                    "ctr": _m.keyword_forecast.ctr,
                    "average_cpc": _m.keyword_forecast.average_cpc,
                    "cost_micros": _m.keyword_forecast.cost_micros}
                   for _m in response_forcast.keyword_forecasts]

        return metrics

    def generate_historical_metrics(self,
                                    keyword_plan_resource_name: str,
                                    stripped_keyword_dict: dict):

        response_historical = self.keyword_plan_service.generate_historical_metrics(
            keyword_plan=keyword_plan_resource_name
        )

        _date_today = datetime.date.today()
        metrics = [{"date_obtained": _date_today,
                    "query": _m.search_query,
                    "avg_searches": _m.keyword_metrics.avg_monthly_searches,
                    "competition": _m.keyword_metrics.competition.name,
                    "competition_index": _m.keyword_metrics.competition_index,
                    "low_top_of_page_bid_micros": _m.keyword_metrics.low_top_of_page_bid_micros,
                    "high_top_of_page_bid_micros": _m.keyword_metrics.high_top_of_page_bid_micros,
                    "volume_trend": [v.monthly_searches for v in _m.keyword_metrics.monthly_search_volumes],
                    "volume_trend_tuples": [(v.monthly_searches, v.year, v.month.value, v.month.name)
                                            for v in _m.keyword_metrics.monthly_search_volumes]}
                   for _m in response_historical.metrics]

        empty_metric = {"date_obtained": _date_today,
                        "query": "",
                        "avg_searches": 0,
                        "competition": "",
                        "competition_index": None,
                        "low_top_of_page_bid_micros": 0,
                        "high_top_of_page_bid_micros": 0,
                        "volume_trend": [],
                        "volume_trend_tuples": []}

        metrics_dict = {d.get("query"): d for d in metrics}

        metrics = []
        for keyword, query in stripped_keyword_dict.items():
            d = metrics_dict.get(query)
            if d:
                # These are queries that were a result of stripping an input keyword (usually equal to the keyword)
                # and also found in the metrics returned by `generate_historical_metrics`.
                d["keyword"] = keyword
                d["status"] = "OBTAINED"
            else:
                # These are queries that were a result of stripping an input keyword (usually equal to the keyword)
                # but are NOT found in the metrics returned by `generate_historical_metrics`.
                d = empty_metric
                d["keyword"] = keyword
                d["query"] = query
                d["status"] = "NOT OBTAINED"
            metrics.append(d)

        for query, d in metrics_dict.items():
            if query not in stripped_keyword_dict.values():
                # These are queries that are found in the metrics returned by `generate_historical_metrics`
                # but were not in the stripped keywords. Keyword planner is stripping some of them further
                # based on its own rules that we don't quite understand.
                d["keyword"] = ""
                d["status"] = "ADDITIONAL"
                metrics.append(d)

        return metrics

    def add_keyword_plan(self,
                         keywords: list[str] | str,
                         match_type: str = "EXACT",
                         cpc_bid_micros: list[int] | int = 10000,
                         forecast_interval: str = "NEXT_QUARTER",
                         location_codes: list[str] = None,
                         language_id: str = None) -> tuple[str, dict]:
        """
        Adds a keyword plan, campaign, ad group, etc. to the customer account.

        Returns: keyword plan resource name
        """
        if location_codes is None:
            location_resource_names = self.location_rns
        else:
            location_resource_names = _map_location_to_resource_names(client=self.client,
                                                                      location_codes=location_codes)

        if language_id is None:
            language_rn = self.language_rn
        else:
            language_rn = _map_language_id_to_resource_name(client=self.client,
                                                            language_id=language_id)

        if isinstance(keywords, str):
            keywords = [keywords]

        # It is necessary to strip any punctuation and non-ascii chars from the keyword before adding to a keyword plan
        # This method therefore provides a stripped_keyword_dict which maps keywords to their "stripped" counterparts
        # this is many-to-one mapping since two multiple different keywords may be mapped to the same stripped
        # counterpart, e.g.:
        # stripped_keyword_dict = {"i love chocolate!!":  "i love chocolate",
        #                          "i-love-chocolate??": "i love chocolate"}
        stripped_keyword_dict = {kw: strip_illegal_chars(kw, language_id) for kw in keywords}
        keywords = [kw for kw in set(stripped_keyword_dict.values())
                    if 0 < len(kw) <= 70 and len(kw.split(' ')) <= 10]

        if isinstance(cpc_bid_micros, int):
            cpc_bid_micros = [cpc_bid_micros] * len(keywords)

        keyword_plan: str = self._create_keyword_plan(forecast_interval=forecast_interval)
        keyword_plan_campaign = self._create_keyword_plan_campaign(
            keyword_plan=keyword_plan,
            location_resources=location_resource_names,
            language_resource=language_rn
        )
        keyword_plan_ad_group = self._create_keyword_plan_ad_group(
            keyword_plan_campaign=keyword_plan_campaign
        )
        keyword_plan_keywords = self._create_keyword_plan_ad_group_keywords(
            keyword_plan_ad_group=keyword_plan_ad_group,
            keyword_texts=keywords,
            match_type=match_type,
            cpc_bid_micros=cpc_bid_micros
        )
        return keyword_plan, stripped_keyword_dict

    def _create_keyword_plan(self, forecast_interval: str = "NEXT_QUARTER") -> str:
        """Adds a keyword plan to the customer account.

        Returns:
            A str of the resource_name for the newly created keyword plan.
        """

        operation = self.client.get_type("KeywordPlanOperation")
        keyword_plan = operation.create

        keyword_plan.name = f"Keyword plan for traffic estimate {uuid.uuid4()}"

        if forecast_interval == "NEXT_QUARTER":
            forecast_interval_enum = (
                self.client.enums.KeywordPlanForecastIntervalEnum.NEXT_QUARTER
            )
        else:  # Defaults to NEXT_QUARTER
            forecast_interval_enum = (
                self.client.enums.KeywordPlanForecastIntervalEnum.NEXT_QUARTER
            )
        keyword_plan.forecast_period.date_interval = forecast_interval_enum

        response = self.keyword_plan_service.mutate_keyword_plans(
            customer_id=self.customer_id, operations=[operation]
        )
        resource_name = response.results[0].resource_name
        # print(f"Created keyword plan with resource name: {resource_name}")
        return resource_name

    def _create_keyword_plan_campaign(self,
                                      keyword_plan,
                                      location_resources,
                                      language_resource,
                                      network=None,
                                      cpc_bid_micros: int = None) -> str:
        """Adds a keyword plan campaign to the given keyword plan.

        Args:
            keyword_plan: A str of the keyword plan resource_name this keyword plan
                campaign should be attributed to.create_keyword_plan.

        Returns:
            A str of the resource_name for the newly created keyword plan campaign.
        """
        if network is None:
            network = self.default_network
        if cpc_bid_micros is None:
            cpc_bid_micros = self.default_keyword_plan_campaign_cpc_bid_micros

        operation = self.client.get_type("KeywordPlanCampaignOperation", version="v15")
        keyword_plan_campaign = operation.create

        keyword_plan_campaign.name = f"Keyword plan campaign {uuid.uuid4()}"
        keyword_plan_campaign.cpc_bid_micros = cpc_bid_micros
        keyword_plan_campaign.keyword_plan = keyword_plan
        keyword_plan_campaign.keyword_plan_network = network

        # Other geotarget constants can be referenced here:
        # https://developers.google.com/google-ads/api/reference/data/geotargets
        for _loc in location_resources:
            geo_target = self.client.get_type("KeywordPlanGeoTarget", version="v15")
            # print('geo_target type: ')
            # print(type(geo_target))
            geo_target.geo_target_constant = _loc
            keyword_plan_campaign.geo_targets.append(geo_target)

        keyword_plan_campaign.language_constants.append(language_resource)

        response = self.keyword_plan_campaign_service.mutate_keyword_plan_campaigns(
            customer_id=self.customer_id, operations=[operation]
        )
        resource_name = response.results[0].resource_name

        return resource_name

    def _create_keyword_plan_ad_group(self,
                                      keyword_plan_campaign: str,
                                      cpc_bid_micros: int = None) -> str:
        """Adds a keyword plan ad group to the given keyword plan campaign.

        Args:
            keyword_plan_campaign: A str of the keyword plan campaign resource_name
                this keyword plan ad group should be attributed to.

        Returns:
            A str of the resource_name for the newly created keyword plan ad group.
        """
        if cpc_bid_micros is None:
            cpc_bid_micros = self.default_keyword_plan_ad_group_cpc_bid_micros

        operation = self.client.get_type("KeywordPlanAdGroupOperation")
        keyword_plan_ad_group = operation.create

        keyword_plan_ad_group.name = f"Keyword plan ad group {uuid.uuid4()}"
        keyword_plan_ad_group.cpc_bid_micros = cpc_bid_micros
        keyword_plan_ad_group.keyword_plan_campaign = keyword_plan_campaign

        response = self.keyword_plan_ad_group_service.mutate_keyword_plan_ad_groups(
            customer_id=self.customer_id, operations=[operation]
        )
        resource_name = response.results[0].resource_name

        return resource_name

    def _create_keyword_plan_ad_group_keywords(self,
                                               keyword_plan_ad_group,
                                               match_type: str,
                                               keyword_texts: list[str],
                                               cpc_bid_micros: list[int]):
        """Adds keyword plan ad group keywords to the given keyword plan ad group.

        Args:
            keyword_plan_ad_group: A str of the keyword plan ad group resource_name
                these keyword plan keywords should be attributed to.

        Raises:
            GoogleAdsException: If an error is returned from the API.
        """

        operations = []

        for keyword_text, cpc_bid in zip(keyword_texts, cpc_bid_micros):
            operation = self.client.get_type("KeywordPlanAdGroupKeywordOperation")
            keyword_plan_ad_group_keyword = operation.create
            keyword_plan_ad_group_keyword.text = keyword_text
            keyword_plan_ad_group_keyword.cpc_bid_micros = cpc_bid
            if match_type == "EXACT":
                keyword_plan_ad_group_keyword.match_type = (
                    self.client.enums.KeywordMatchTypeEnum.EXACT
                )
            elif match_type == "BROAD":
                keyword_plan_ad_group_keyword.match_type = (
                    self.client.enums.KeywordMatchTypeEnum.BROAD
                )
            elif match_type == "PHRASE":
                keyword_plan_ad_group_keyword.match_type = (
                    self.client.enums.KeywordMatchTypeEnum.PHRASE
                )
            keyword_plan_ad_group_keyword.keyword_plan_ad_group = keyword_plan_ad_group

            operations.append(operation)

        response = self.keyword_plan_ad_group_keyword_service.mutate_keyword_plan_ad_group_keywords(
            customer_id=self.customer_id, operations=operations
        )

        # return [result.resource_name for result in response.results]
        return [result for result in response.results]


class KeywordPlanIdeaService(ClientWrapper):
    def __init__(self,
                 googleads_client: GoogleAdsClient,
                 customer_id: str,
                 site_url: str = None,
                 location_codes: list[str] = None,
                 language_id: str | None = None
                 ):
        super().__init__(googleads_client=googleads_client,
                         customer_id=customer_id,
                         location_codes=location_codes,
                         language_id=language_id)

        self.site_url = site_url

        self.keyword_plan_idea_service = self.client.get_service(
            "KeywordPlanIdeaService")
        self.keyword_competition_level_enum = (
            self.client.enums.KeywordPlanCompetitionLevelEnum)
        self.keyword_plan_network = (
            self.client.enums.KeywordPlanNetworkEnum.GOOGLE_SEARCH_AND_PARTNERS
        )

    def generate_keyword_ideas(self,
                               url: str | None = None,
                               phrases: str | list[str] | None = None,
                               include_adult_keywords: bool = True,
                               location_codes: list[str] = None,
                               language_id: str = None) -> pd.DataFrame:
        """
        Use the Google Ads Keyword Planner API to generate keyword ideas based on a URL or a phrase
        """

        if location_codes is None:
            location_resource_names = self.location_rns
        else:
            location_resource_names = _map_location_to_resource_names(client=self.client,
                                                                      location_codes=location_codes)

        if language_id is None:
            language_rn = self.language_rn
        else:
            language_rn = _map_language_id_to_resource_name(client=self.client,
                                                            language_id=language_id)

        request = self.client.get_type("GenerateKeywordIdeasRequest")
        request.customer_id = self.customer_id
        request.language = language_rn
        request.geo_target_constants = location_resource_names
        request.include_adult_keywords = include_adult_keywords
        request.keyword_plan_network = self.keyword_plan_network
        # request.page_size = page_size

        if url is not None:
            if RE_URL.match(url):
                page_url = url
            elif url == "/" or url == "":
                page_url = self.site_url
            elif url[0] == "/":
                page_url = self.site_url + url[1:]
            else:
                page_url = self.site_url + url
        else:
            page_url = None

        # To generate keyword ideas with only a page_url and no keywords we need
        # to initialize a UrlSeed object with the page_url as the "url" field.
        if not phrases and page_url:
            request.url_seed.url = page_url

        # To generate keyword ideas with only a list of keywords and no page_url
        # we need to initialize a KeywordSeed object and set the "keywords" field
        # to be a list of StringValue objects.
        if phrases and not page_url:
            request.keyword_seed.keywords.extend(phrases)

        # To generate keyword ideas using both a list of keywords and a page_url we
        # need to initialize a KeywordAndUrlSeed object, setting both the "url" and
        # "keywords" fields.
        if phrases and page_url:
            request.keyword_and_url_seed.url = page_url
            request.keyword_and_url_seed.keywords.extend(phrases)

        if not phrases and not page_url:
            raise ValueError("must supply either `phrases` or `url`")

        keyword_ideas_pager = self.keyword_plan_idea_service.generate_keyword_ideas(request=request)

        # keyword_ideas_list = [idea for idea in keyword_ideas_pager]

        _date_today = datetime.date.today()
        keyword_ideas = [{"date_obtained": _date_today,
                          "phrase": idea.text,
                          "avg_searches": idea.keyword_idea_metrics.avg_monthly_searches,
                          "competition": idea.keyword_idea_metrics.competition.name,
                          "competition_index": idea.keyword_idea_metrics.competition_index,
                          "low_top_of_page_bid_micros": idea.keyword_idea_metrics.low_top_of_page_bid_micros,
                          "high_top_of_page_bid_micros": idea.keyword_idea_metrics.high_top_of_page_bid_micros,
                          "volume_trend": [v.monthly_searches
                                           for v in idea.keyword_idea_metrics.monthly_search_volumes]
                          }
                         for idea in keyword_ideas_pager]

        _df = pd.DataFrame(keyword_ideas)
        _df["volume_trend_coef"] = _df["volume_trend"].apply(_three_month_trend_coef)
        _df["latest_volume"] = _df["volume_trend"].apply(_latest_volume)

        return _df

    def generate_historical_metrics(self,
                                   keywords: str | list[str] | None = None,
                                   include_adult_keywords: bool = True,
                                   location_codes: list[str] = None,
                                   language_id: str = None,
                                    calculated_fields: bool = True,
                                    concat_locations: bool = True,
                                    print_progress: bool = False) -> tuple[pd.DataFrame, pd.DataFrame]:
        """
        Use the Google Ads Keyword Planner API to generate keyword ideas based on a URL or a phrase
        """
        if isinstance(keywords, str):
            keywords = [keywords]

        if language_id is None:
            language_rn = self.language_rn
        else:
            language_rn = _map_language_id_to_resource_name(client=self.client,
                                                            language_id=language_id)

        keyword_list_partition = _partition_list(_list=keywords, n=2_000)
        frames: list[pd.DataFrame] = []

        for _sublist_index, _keyword_sublist in enumerate(keyword_list_partition):
            if _sublist_index > 0:
                time.sleep(API_MULTIPLE_REQUEST_WAIT_TIME)

            try:
                historical_metrics_df = self._historical_metrics_dataframe(
                    keywords=_keyword_sublist,
                    location_codes=location_codes,
                    language_resource_name=language_rn,
                    include_adult_keywords=include_adult_keywords,
                    calculated_fields=calculated_fields,
                    concat_locations=concat_locations
                )

                if historical_metrics_df is not None:
                    frames.append(historical_metrics_df)

            except GoogleAdsException:
                print(f"GoogleAdsException encountered after obtaining {len(frames)} frames")
                break

            if print_progress:
                print(f"{len(frames)} - obtained metrics for {len(frames[-1])} keywords")

        if len(frames) > 0:
            _df = pd.concat(frames)
        else:
            return None, None

        metrics_df: pd.DataFrame = _df.drop(columns=['volume_trend_tuples'])

        monthly_volume_df: pd.DataFrame = _df[['country_iso_code',
                                               'date_obtained',
                                               'query',
                                               'keyword',
                                               'status',
                                               'volume_trend_tuples']].explode('volume_trend_tuples')

        monthly_volume_df['volume_trend_tuples'] = monthly_volume_df['volume_trend_tuples'].apply(
            _check_volume_trend_tuples)

        monthly_volume_df['month_name'] = monthly_volume_df['volume_trend_tuples'].apply(lambda t: t[3])
        monthly_volume_df['month_enum'] = monthly_volume_df['volume_trend_tuples'].apply(lambda t: t[2])
        monthly_volume_df['month'] = monthly_volume_df['month_enum'].apply(_get_month_from_enum_value)
        monthly_volume_df['year'] = monthly_volume_df['volume_trend_tuples'].apply(lambda t: t[1])

        monthly_volume_df.dropna(axis='index', subset=['month'], inplace=True)

        monthly_volume_df['record_date'] = monthly_volume_df.apply(
            lambda df: _conv_date(df['year'], df['month'], 15, month_index=0),
            axis=1
        )
        monthly_volume_df['volume'] = monthly_volume_df['volume_trend_tuples'].apply(lambda t: t[0])

        monthly_volume_df.drop(
            columns=['month_enum', 'volume_trend_tuples', 'month', 'year', 'month_name'],
            inplace=True
        )

        metrics_df.drop_duplicates(subset=["country_iso_code", "query", "keyword"], keep="first", inplace=True)
        monthly_volume_df.drop_duplicates(subset=["country_iso_code", "query", "keyword", "record_date"], keep="first", inplace=True)

        metrics_df.reset_index(drop=True, inplace=True)
        monthly_volume_df.reset_index(drop=True, inplace=True)

        _recent_dates = monthly_volume_df[
            ["country_iso_code", 'query', 'keyword', 'record_date']
        ].groupby(by=["country_iso_code", 'query', 'keyword']).max().reset_index()

        _competition_subset = pd.merge(
            left=_recent_dates,
            right=metrics_df[["country_iso_code", 'query', 'keyword', 'competition', 'competition_index']],
            on=["country_iso_code", 'query', 'keyword'],
            how='inner'
        )

        metrics_df = pd.merge(left=metrics_df, right=_recent_dates, on=["country_iso_code", 'query', 'keyword'], how='left')

        monthly_volume_df = pd.merge(
            left=monthly_volume_df,
            right=_competition_subset,
            on=["country_iso_code", 'query', 'keyword', 'record_date'],
            how='left'
        )

        return metrics_df, monthly_volume_df


    def _historical_metrics_response(self,
                                  keywords: list[str],
                                  location_resource_names: list,
                                  language_resource_name: str,
                                  include_adult_keywords: bool = True):
        request = self.client.get_type("GenerateKeywordHistoricalMetricsRequest")
        request.customer_id = self.customer_id
        request.language = language_resource_name
        request.geo_target_constants = location_resource_names
        request.include_adult_keywords = include_adult_keywords
        request.keyword_plan_network = self.keyword_plan_network
        request.keywords = keywords

        return self.keyword_plan_idea_service.generate_keyword_historical_metrics(
            request=request
        )

    def _historical_metrics_dataframe(self,
                                  keywords: list[str],
                                  location_codes: list,
                                  language_resource_name: str,
                                  include_adult_keywords: bool = True,
                                      calculated_fields: bool = True,
                                      concat_locations: bool = True):

        if location_codes is None:
            location_codes = self.location_codes
            location_resource_names = self.location_rns
        else:
            location_resource_names = _map_location_to_resource_names(client=self.client,
                                                                      location_codes=location_codes)

        if concat_locations:
            return self._historical_metrics_dataframe_concat_locations(
                keywords=keywords,
                location_codes=location_codes,
                language_resource_name=language_resource_name,
                include_adult_keywords=include_adult_keywords,
                calculated_fields=calculated_fields
            )

        response = self._historical_metrics_response(
            keywords=keywords,
            location_resource_names=location_resource_names,
            language_resource_name=language_resource_name,
            include_adult_keywords=include_adult_keywords
        )

        _date_today = datetime.date.today()

        metrics = []
        for _m in response.results:
            _query= _m.text
            _vars = _m.close_variants
            if not _vars:
                _vars = []
            _keywords = [_query] + _vars

            for _k in _keywords:

                _metric = {
                    "date_obtained": _date_today,
                    "country_iso_code": '-'.join(location_codes),
                    "query": _query,
                    "keyword": _k,
                    "status": "OBTAINED",
                    "avg_searches": _m.keyword_metrics.avg_monthly_searches,
                    "competition": _m.keyword_metrics.competition.name,
                    "competition_index": _m.keyword_metrics.competition_index,
                    "low_top_of_page_bid_micros": _m.keyword_metrics.low_top_of_page_bid_micros,
                    "high_top_of_page_bid_micros": _m.keyword_metrics.high_top_of_page_bid_micros,
                    "volume_trend": [v.monthly_searches for v in _m.keyword_metrics.monthly_search_volumes],
                    "volume_trend_tuples": [(v.monthly_searches, v.year, v.month.value, v.month.name)
                                                for v in _m.keyword_metrics.monthly_search_volumes]
                }
                metrics.append(_metric)

        empty_metric = {
            "date_obtained": _date_today,
            "country_iso_code": '-'.join(location_codes),
            "query": "",
            "status": "NOT_OBTAINED",
            "avg_searches": 0,
            "competition": "",
            "competition_index": None,
            "low_top_of_page_bid_micros": 0,
            "high_top_of_page_bid_micros": 0,
            "volume_trend": [],
            "volume_trend_tuples": []
        }

        for _missing_keyword in set(keywords) - set(m.get("keyword") for m in metrics):
            _metric = empty_metric.copy()
            _metric["keyword"] = _missing_keyword
            metrics.append(_metric)

        if len(metrics) == 0:
            df = pd.DataFrame(
                columns=[
                    "date_obtained",
                    "country_iso_code",
                    "query",
                    "keyword",
                    "status",
                    "avg_searches",
                    "competition",
                    "competition_index",
                    "low_top_of_page_bid_micros",
                    "high_top_of_page_bid_micros",
                    "volume_trend",
                    "volume_trend_tuples"
                ]
            )
        else:
            df = pd.DataFrame(metrics)

        if calculated_fields:
            df["volume_trend_coef"] = df["volume_trend"].apply(_three_month_trend_coef)
            df["volume"] = df["volume_trend"].apply(_latest_volume)
            df["volume_last_month"] = df["volume_trend"].apply(_previous_volume)
            df["volume_last_year"] = df["volume_trend"].apply(_last_year_volume)
            df["volume_3monthavg"] = df["volume_trend"].apply(_volume_3monthavg)
            df["volume_6monthavg"] = df["volume_trend"].apply(_volume_6monthavg)
            df["volume_12monthavg"] = df["volume_trend"].apply(_volume_12monthavg)

        return df

    def _historical_metrics_dataframe_concat_locations(self,
                                                      keywords: list[str],
                                                      location_codes: list[str],
                                                      language_resource_name: str,
                                                      include_adult_keywords: bool = True,
                                                       calculated_fields: bool = True):
        frames = []
        for _loc in location_codes:
            _df = self._historical_metrics_dataframe(
                keywords=keywords,
                location_codes=[_loc],
                language_resource_name=language_resource_name,
                include_adult_keywords=include_adult_keywords,
                calculated_fields=calculated_fields,
                concat_locations=False
            )
            if len(_df) > 0:
                frames.append(_df)

        if len(frames) > 0:
            df = pd.concat(frames)
        else:
            df = pd.DataFrame(
                columns=[
                    "date_obtained",
                    "country_iso_code",
                    "query",
                    "keyword",
                    "status",
                    "avg_searches",
                    "competition",
                    "competition_index",
                    "low_top_of_page_bid_micros",
                    "high_top_of_page_bid_micros",
                    "volume_trend",
                    "volume_trend_tuples"
                ]
            )

        return df



def _three_month_trend_coef(volume_trends: list[int]):
    if not isinstance(volume_trends, list):
        return None
    if len(volume_trends) == 0:
        return None
    y = np.array(volume_trends)
    x = np.arange(len(y))
    p = np.polyfit(x[-3:], y[-3:], deg=1)
    return round(p[0] / y.mean(), 3)


def _latest_volume(volume_trends):
    if not isinstance(volume_trends, list):
        return None
    if len(volume_trends) == 0:
        return None
    return volume_trends[-1]


def _previous_volume(volume_trends):
    if not isinstance(volume_trends, list):
        return None
    if len(volume_trends) == 0:
        return None
    return volume_trends[-2]


def _volume_mom(volume_trends):
    if not isinstance(volume_trends, list):
        return None
    if len(volume_trends) < 2:
        return None
    return volume_trends[-1] - volume_trends[-2]


def _last_year_volume(volume_trends):
    if not isinstance(volume_trends, list):
        return None
    if len(volume_trends) < 12:
        return None
    return volume_trends[-12]


def _volume_yoy(volume_trends):
    if not isinstance(volume_trends, list):
        return None
    if len(volume_trends) < 12:
        return None
    return volume_trends[-1] - volume_trends[-12]


def _volume_3monthavg(volume_trends):
    if not isinstance(volume_trends, list):
        return None
    if len(volume_trends) == 0:
        return None
    return np.mean(volume_trends[-3:])


def _volume_6monthavg(volume_trends):
    if not isinstance(volume_trends, list):
        return None
    if len(volume_trends) == 0:
        return None
    return np.mean(volume_trends[-6:])


def _volume_12monthavg(volume_trends):
    if not isinstance(volume_trends, list):
        return None
    if len(volume_trends) == 0:
        return None
    return np.mean(volume_trends[-12:])


def _partition_list(_list, n):
    return [_list[n * i:n * i + n] for i in range(ceil(len(_list) / n))]


def _conv_date(year, month, day, month_index: int = 0):
    if isinstance(year, float):
        year = int(year)
    if not isinstance(year, int):
        return None
    if isinstance(month, float):
        month = int(month)
    if not isinstance(month, int):
        return None
    if month_index == 0:
        month += 1
    return datetime.date(year, month, day)


def _map_location_to_resource_names(client,
                                    location_codes: str | list[str]):
    if isinstance(location_codes, str):
        location_codes = [location_codes]

    location_ids = []
    for _code in location_codes:
        _code = str(_code)
        if re.match("[A-Z]{3}", _code):
            _id = LOCATION_ID_DICT.get(_code, None)
            if _id:
                location_ids.append(_id)
        elif re.match("[0-9]+", _code):
            location_ids.append(_code)

    build_resource_name = client.get_service("GeoTargetConstantService").geo_target_constant_path
    return [build_resource_name(location_id) for location_id in location_ids]


def _map_locations_ids_to_resource_names(client, location_ids):
    """Converts a list of location IDs to resource names.

    Args:
        client: an initialized GoogleAdsClient instance.
        location_ids: a list of location ID strings.

    Returns:
        a list of resource name strings using the given location IDs.
    """
    build_resource_name = client.get_service("GeoTargetConstantService").geo_target_constant_path
    return [build_resource_name(location_id) for location_id in location_ids]


def _map_language_id_to_resource_name(client, language_id):
    return client.get_service("GoogleAdsService").language_constant_path(criterion_id=language_id)


def _check_volume_trend_tuples(t):
    if isinstance(t, tuple) and len(t) == 4:
        return t
    else:
        return 0, 0, 0, ''


def _get_month_from_enum_value(month_enum_value):
    if month_enum_value <= 1:
        return None
    else:
        return month_enum_value - 2


def strip_illegal_chars(s, language_id):
    # lower case obvs
    s = s.lower()
    # replace `‘’ with simple '
    s = re.sub(r"[`‘’]+", "'", s)
    # first remove any non-ascii characters, replace with a space
    if language_id in ['1001']:  # when german language_id don't replace one of [Ä, ä, Ö, ö, Ü, ü, ß]
        s = re.sub(r"[^\x00-\x7F\xC4\xE4\xD6\xF6\xDC\xFC\xDF]+", " ", s)
    else:
        s = re.sub(r"[^\x00-\x7F]+", " ", s)
    # then replace any punctuation with a space
    s = re.sub(r"""[!¡⁄@%^()={}:;,.+\-–—_±~“"#<>?/|*¶•ªº&\[\]\\]+""", " ", s)
    # finally, replace multiple spaces (r"\s+") with a single space
    s = re.sub(r"\s+", " ", s)
    return s.strip()



import logging
import re
import csv
import os
import numpy as np
import pandas as pd
import datetime

from google.analytics.data_v1beta.types.analytics_data_api import RunReportResponse

from typing import List, Optional, Union, Pattern

from .utils import general_utils
from .utils.ga4_parser import parse_ga4_response, parse_ga3_response






from typing import List

from google.ads.googleads.client import GoogleAdsClient

from pygoogalytics.utils.resource_utils import get_analytics_resources, \
    googleads_client_from_yaml, googleads_client_from_key_file, parse_ads_id
from .googalytics_wrapper import GoogalyticsWrapper
from .kwp_wrappers import KeywordPlanIdeaService, KeywordPlanService
from .ads_wrapper import AdsWrapper
from . import pga_logger


class GoogalyticsClient:
    """
    The Client class holds the credentials for a project
    which can then be used to create a GoogalyticsWrapper object.

    example implementation:
    from pygoogalytics.client import Client
    googalytics_client = Client(key_file_path='<path-to-your-key-file>')
    """

    def __init__(self,
                 gsc_resource=None,
                 ga3_resource=None,
                 ga4_resource=None
                 ):

        self.gsc_resource = gsc_resource
        self.ga3_resource = ga3_resource
        self.ga4_resource = ga4_resource

    @classmethod
    def build(cls, api_key: str = None, key_file_path: str = None):
        _ga3_resource, _ga4_resource, _gsc_resource = get_analytics_resources(
            json_api_key=api_key, key_file_path=key_file_path
        )
        return cls(gsc_resource=_gsc_resource, ga3_resource=_ga3_resource,  ga4_resource=_ga4_resource)

    def wrapper(self,
                sc_domain: str = None,
                view_id: str = None,
                ga4_property_id: str = None,
                ) -> GoogalyticsWrapper:
        """
        The GoogalyticsWrapper requires different arguments to access data depending on the source.
        @param sc_domain: required for GSC data. This is the url-like string you see in the Google Search Console web application when selecting the site. It is either a full url (e.g. `https://www.example.com/`) or something like `sc_domain:example.com`
        @param view_id: required for GA3 data: the "view_id" you see in "settings" on the GA web application. This is usually an 8- or 9-digit number, passed as a string
        @param ga4_property_id: required for GA4 data. Note that GA4 is currently not fully supported by PyGoogalytics
        @returns: GoogalyticsWrapper object.
        """
        return GoogalyticsWrapper(
            gsc_resource=self.gsc_resource,
            ga3_resource=self.ga3_resource,
            ga4_resource=self.ga4_resource,
            sc_domain=sc_domain,
            view_id=view_id,
            ga4_property_id=ga4_property_id
        )

    def __bool__(self):
        if self.gsc_resource is not None or self.ga3_resource is not None or self.ga4_resource is not None:
            return True
        else:
            return False

    def __dict__(self):
        return {
            'gsc_resource': self.gsc_resource.__repr__(),
            'ga3_resource': self.ga3_resource.__repr__(),
            'ga4_resource': self.ga4_resource.__repr__()
        }

    def __repr__(self):
        _s = 'PyGoogalytics Client object:\n'
        for _k, _v in self.__dict__().items():
            _s += f" - {_k}: {_v}\n"
        return _s


class AdsClient:
    def __init__(self,
                 googleads_client: GoogleAdsClient,
                 default_customer_id: str = None,
                 googleads_yaml_dict: dict = None):
        self.googleads_client = googleads_client
        self.default_customer_id = default_customer_id
        self.googleads_yaml_dict = googleads_yaml_dict

    @classmethod
    def build(cls, yaml_key: str = None, key_file_path: str = None):
        if yaml_key:
            googleads_client, default_customer_id, googleads_yaml_dict = googleads_client_from_yaml(googleads_yaml_string=yaml_key)
            pga_logger.info(f"initialised KwpClient object from yaml string")
        elif key_file_path:
            googleads_client, default_customer_id, googleads_yaml_dict = googleads_client_from_key_file(path=key_file_path)
            pga_logger.info(f"initialised KwpClient object from key file")
        else:
            raise KeyError("either yaml_key or key_file_path must be supplied")

        return cls(googleads_client=googleads_client,
                   default_customer_id=default_customer_id,
                   googleads_yaml_dict=googleads_yaml_dict)

    def __bool__(self):
        if self.googleads_client is not None:
            return True
        else:
            return False

    def plan_service(self,
                     customer_id: str = None,
                     location_codes: List[str] = None,
                     language_id: str = None
                     ):
        if customer_id is None:
            customer_id = self.default_customer_id

        return KeywordPlanService(googleads_client=self.googleads_client,
                                  customer_id=parse_ads_id(customer_id),
                                  location_codes=location_codes,
                                  language_id=language_id)

    def ideas_service(self,
                      customer_id: str = None,
                      site_url: str = None,
                      location_codes: List[str] = None,
                      language_id: str = None
                      ):
        if customer_id is None:
            customer_id = self.default_customer_id

        return KeywordPlanIdeaService(googleads_client=self.googleads_client,
                                      customer_id=parse_ads_id(customer_id),
                                      site_url=site_url,
                                      location_codes=location_codes,
                                      language_id=language_id)

    def report_service(self, customer_id: str):
        if customer_id is None:
            customer_id = self.default_customer_id
        return AdsWrapper(googleads_client=self.googleads_client,
                          customer_id=parse_ads_id(customer_id))



import logging
import re
import csv
import os
import numpy as np
import pandas as pd
import datetime

from google.analytics.data_v1beta.types.analytics_data_api import RunReportResponse

from typing import List, Optional, Union, Pattern

from .utils import general_utils
from .utils.ga4_parser import parse_ga4_response, parse_ga3_response


def camel_to_snake(string: str):
    return general_utils.RE_C2S.sub('_', string).lower()


def from_response(response: dict | RunReportResponse,
                  response_type: str = None,
                  report_index: int = 0,
                  gsc_dimensions: Optional[List[str]] = None,
                  start_date: datetime.date = None,
                  end_date: datetime.date = None):
    # We're going to what kind of response this is by making assumptions about what the responses will look like.
    # If the response dictionary includes a 'report' key, we assume it is a GA response.
    # If the response dictionary includes a 'rows' key with returns
    if response_type is None:
        response_type = get_response_type(response)

    if response_type == 'GA3' or response_type == 'GA4':
        if isinstance(response, RunReportResponse):
            response = parse_ga4_response(response)
        elif response.get('reports', None):
            response = parse_ga3_response(response)
        return from_standard_analytics_response(response, start_date=start_date, end_date=end_date)

    elif response_type == 'GSC':
        rows = response.get('rows', [])
        response_aggregation = response.get('responseAggregationType', None)
        _gsc_df = GSCDataFrame(df_input=rows,
                               gsc_dimensions=gsc_dimensions,
                               from_gsc_response=True)
        _gsc_df.response_aggregation = response_aggregation
        return _gsc_df

def from_standard_analytics_response(response: dict, start_date: datetime.date=None, end_date: datetime.date=None):
    rows = response.get('rows')
    from_ga_response = True
    if len(rows) == 0:
        rows = None
        from_ga_response = False

    if response.get('start_date'):
        start_date = response.get('start_date')
    if response.get('end_date'):
        end_date = response.get('end_date')

    return GADataFrame(df_input=rows,
                       response_type='GA4',
                       dimensions=response.get('dimension_headers'),
                       metrics=response.get('metric_headers'),
                       row_count=response.get('row_count'),
                       start_date=start_date,
                       end_date=end_date,
                       error=response.get('error_type'),
                       from_ga_response=from_ga_response)

def from_csv(csv_file_path):
    pandas_df = pd.read_csv(csv_file_path)
    r_type = response_type_from_file(csv_file_path)
    if r_type == 'GA':
        return GADataFrame(df_input=pandas_df,
                           from_ga_response=False)
    if r_type == 'GSC' or r_type == 'GSCQ':
        return GSCDataFrame(df_input=pandas_df,
                            from_gsc_response=False)


class GADataFrame(pd.DataFrame):
    """
    Subclass of pandas.DataFrame specifically for use with Google Analytics responses.
    The subclass must be initialized with a GA response object and will create a dataframe
    from the report indexed by `report_index` (default 0).

    In addition to pandas DataFrame functionality, the GADataFrame has additional metadata attributes:
        dimensions:      list of the GA dimensions returned in the response
        metrics:         list of the GA metrics returned in the response
        brand_regex_dict:     dictionary of regex patterns supplied to __init__()
        category_regex_dict: dictionary of regex patterns supplied to __init__()
    """

    _metadata = ["dimensions",
                 "metrics",
                 "join_dimensions",
                 "date_range",
                 "date_range_days",
                 "time_obtained",
                 "row_count",
                 "response_type",
                 "error"]

    def __init__(self, df_input,
                 dimensions: list[str] = None,
                 metrics: list[str] = None,
                 from_ga_response: bool = False,
                 join_dimensions: List[str] = None,
                 start_date: datetime.date = None,
                 end_date: datetime.date = None,
                 time_obtained: datetime.datetime = None,
                 row_count: int = 0,
                 response_type: str = 'GA4',
                 error: str | None = None
                 ):

        if time_obtained:
            self.time_obtained = time_obtained.astimezone(tz=datetime.tzinfo())
        else:
            self.time_obtained = datetime.datetime.now(tz=datetime.timezone.utc)

        self.dimensions = dimensions
        self.metrics = metrics

        if join_dimensions is None:
            join_dimensions = dimensions

        self.join_dimensions = join_dimensions

        self.date_range = {
            'start': start_date or datetime.date.today(),
            'end': end_date or datetime.date.today()
        }
        self.row_count = row_count
        self.response_type = response_type
        self.error = error

        if start_date and end_date:
            self.date_range_days = (end_date - start_date).days + 1
        else:
            self.date_range_days = None

        if df_input is None:
            self.join_dimensions = self.dimensions
            super().__init__(None, columns=dimensions + metrics)

            if 'landingPagePlusQueryString' in self.columns:
                self.rename(columns={'landingPagePlusQueryString': 'landingPagePath'}, inplace=True)
                self.remove_join_dimensions('landingPagePlusQueryString')
                self.add_join_dimensions('landingPagePath')

            if 'landingPagePath' in self.columns:
                # rename 'landingPagePath' to 'landingPageFull', then add dummy columns for
                # 'landingPage' and 'landingPageParameter'
                self.rename(columns={'landingPagePath': 'landingPageFull'}, inplace=True)
                self['landingPage'] = None
                self['landingPageParameter'] = None
                self.remove_join_dimensions('landingPagePath')
                self.add_join_dimensions(['landingPage', 'landingPageFull', 'landingPageParameter'])

            if 'countryId' in self.columns:
                self.rename(columns={'countryId': 'countryIsoCode'}, inplace=True)
                self.remove_join_dimensions('countryId')
                self.add_join_dimensions('countryIsoCode')

            if 'deviceCategory' in self.columns:
                self.rename(columns={'deviceCategory': 'device'}, inplace=True)
                self.remove_join_dimensions('deviceCategory')
                self.add_join_dimensions('device')

            if 'sourceMedium' in self.columns:
                self['source'] = None
                self['medium'] = None
                self.drop(columns='sourceMedium', inplace=True)
                self.remove_join_dimensions('sourceMedium')
                self.add_join_dimensions(['source', 'medium'])

            if 'sessionSourceMedium' in self.columns:
                self['source'] = None
                self['medium'] = None
                self.drop(columns='sessionSourceMedium', inplace=True)
                self.remove_join_dimensions('sessionSourceMedium')
                self.add_join_dimensions(['source', 'medium'])

            if 'transactionsPerSession' in self.columns:
                self.rename(columns={'transactionsPerSession': 'conversionRate'}, inplace=True)

            if 'itemPurchaseQuantity' in self.columns:
                self.rename(columns={'itemPurchaseQuantity': 'itemQuantity'}, inplace=True)
                self.metrics = remove_list_item(self.metrics, 'itemPurchaseQuantity')
                self.metrics.append('itemQuantity')

            if 'date' in self.columns:
                self.rename(columns={'date': 'recordDate'}, inplace=True)
                self.remove_join_dimensions('date')
                self.add_join_dimensions(['recordDate'])

            if 'dateHourMinute' in self.columns:
                self['recordDate'] = None
                self['recordTime'] = None
                self.drop(columns='dateHourMinute', inplace=True)
                self.remove_join_dimensions('dateHourMinute')
                self.add_join_dimensions(['recordDate', 'recordTime'])

            if 'dateHour' in self.columns:
                self['recordDate'] = None
                self['recordTime'] = None
                self.drop(columns='dateHour', inplace=True)
                self.remove_join_dimensions('dateHour')
                self.add_join_dimensions(['recordDate', 'recordTime'])

            self.snake_case_join_dimensions()

        else:
            super().__init__(df_input)
            for _c in self.columns:
                if _c in self.metrics:
                    self[_c] = self[_c].astype('float')

        if from_ga_response:
            if 'productName' in self.columns:
                self['productName'] = self['productName'].apply(lambda s: s.strip().lower()
                                                                .replace('-', ' ')
                                                                .replace('_', ' ')
                                                                .replace('&amp;', '&')
                                                                )
                self['productName'] = self['productName'].apply(lambda s: re.sub(r"\\u[a-f\d]{4}", " ", s))
                self['productName'] = self['productName'].apply(lambda s: re.sub(r"\s+", " ", s).strip())

            if 'landingPagePlusQueryString' in self.columns:
                self.rename(columns={'landingPagePlusQueryString': 'landingPagePath'}, inplace=True)
                self.remove_join_dimensions('landingPagePlusQueryString')
                self.add_join_dimensions('landingPagePath')

            if 'itemPurchaseQuantity' in self.columns:
                self.rename(columns={'itemPurchaseQuantity': 'itemQuantity'}, inplace=True)
                self.metrics = remove_list_item(self.metrics, 'itemPurchaseQuantity')
                self.metrics.append('itemQuantity')

            if 'landingPagePath' in self.columns:
                self['landingPagePath'] = self['landingPagePath'].apply(general_utils.url_strip_domain)
                self['landingPage'] = self['landingPagePath'].apply(lambda _u: _u.split('?')[0])
                self['landingPageParameter'] = self['landingPagePath'].apply(general_utils.url_extract_parameter)
                self.rename(columns={'landingPagePath': 'landingPageFull'}, inplace=True)

                self.remove_join_dimensions('landingPagePath')
                self.add_join_dimensions(['landingPageFull', 'landingPageParameter', 'landingPage'])

            if 'transactionsPerSession' in self.columns:
                self.rename(columns={'transactionsPerSession': 'conversionRate'}, inplace=True)

            if 'deviceCategory' in self.columns:
                self.rename(columns={'deviceCategory': 'device'}, inplace=True)
                self.remove_join_dimensions('deviceCategory')
                self.add_join_dimensions('device')

            if 'sourceMedium' in self.columns:
                self['source'] = self.sourceMedium.apply(lambda s: s.split('/')[0].strip())
                self['medium'] = self.sourceMedium.apply(lambda s: s.split('/')[1].strip() if '/' in s else None)
                self.drop(columns='sourceMedium', inplace=True)
                self.remove_join_dimensions('sourceMedium')
                self.add_join_dimensions(['source', 'medium'])

            if 'sessionSourceMedium' in self.columns:
                self['source'] = self['sessionSourceMedium'].apply(lambda s: s.split('/')[0].strip())
                self['medium'] = self['sessionSourceMedium'].apply(lambda s: s.split('/')[1].strip() if '/' in s else None)
                self.drop(columns='sessionSourceMedium', inplace=True)
                self.remove_join_dimensions('sessionSourceMedium')
                self.add_join_dimensions(['source', 'medium'])

            if 'date' in self.columns:
                self.drop(
                    self[self.date.apply(lambda _date: True if _date == '(other)' else False)].index,
                    inplace=True
                )
                self.date = pd.to_datetime(self.date, format="%Y%m%d")
                self.date = self.date.apply(lambda date_time: datetime.datetime.date(date_time))
                self.rename(columns={'date': 'recordDate'}, inplace=True)
                self.remove_join_dimensions('date')
                self.add_join_dimensions('recordDate')

            if 'dateHourMinute' in self.columns:
                self.drop(
                    self[self.dateHourMinute.apply(lambda _datetime: True if _datetime == '(other)' else False)].index,
                    inplace=True
                )
                date_time_series = pd.to_datetime(self.dateHourMinute, format="%Y%m%d%H%M")
                self['recordDate'] = date_time_series.apply(lambda date_time: datetime.datetime.date(date_time))
                self['recordTime'] = date_time_series.apply(lambda date_time: date_time.time())
                self.drop(columns='dateHourMinute', inplace=True)
                self.remove_join_dimensions('dateHourMinute')
                self.add_join_dimensions(['recordDate', 'recordTime'])

            if 'dateHour' in self.columns:
                self.drop(
                    self[self['dateHour'].apply(lambda _datetime: True if _datetime == '(other)' else False)].index,
                    inplace=True
                )
                date_time_series = pd.to_datetime(self['dateHour'], format="%Y%m%d%H")
                self['recordDate'] = date_time_series.apply(lambda date_time: datetime.datetime.date(date_time))
                self['recordTime'] = date_time_series.apply(lambda date_time: date_time.time())
                self.drop(columns='dateHour', inplace=True)
                self.remove_join_dimensions('dateHour')
                self.add_join_dimensions(['recordDate', 'recordTime'])

            if 'countryId' in self.columns:
                self.rename(columns={'countryId': 'countryIsoCode'}, inplace=True)
                self.remove_join_dimensions('countryId')
                self.add_join_dimensions(['countryIsoCode'])

            if 'countryIsoCode' in self.columns:
                self.countryIsoCode = self.countryIsoCode.apply(iso_code_2_to_3)

            self.snake_case_join_dimensions()

        self.snake_case_columns()

        if "record_date" not in self.columns:
            if start_date and end_date:
                self.insert(loc=0, column="record_date", value=end_date)
                self.insert(loc=0, column="record_date_start", value=start_date)
                self.add_join_dimensions("record_date")
                self.add_join_dimensions("record_date_start")
            elif end_date:
                self.insert(loc=0, column="record_date", value=end_date)
                self.add_join_dimensions("record_date")
            elif start_date:
                self.insert(loc=0, column="record_date", value=start_date)
                self.add_join_dimensions("record_date")

    def snake_case_columns(self):
        self.rename(columns={old_col_name: camel_to_snake(old_col_name)
                             for old_col_name in self.columns},
                    inplace=True)
        
    def snake_case_join_dimensions(self):
        self.join_dimensions = [camel_to_snake(_) for _ in self.join_dimensions]

    def add_brand_column(self,
                         brand_function: Optional[str],
                         brands_dict: Optional[dict] = None):
        if 'product_name' in self.columns:
            _loc = self.columns.get_loc('product_name') + 1
            if brand_function == "first word":
                self.insert(loc=_loc, column='brand',
                            value=self.product_name.apply(lambda s: s.split(' ')[0]))
            elif brand_function == "regex":
                def regex_brand_function_local(_product_name):
                    return regex_brand_function(product_name=_product_name, brands_dict=brands_dict)

                self.insert(loc=_loc, column='brand',
                            value=self.product_name.apply(regex_brand_function_local))
            else:
                self.insert(loc=_loc, column='brand',
                            value='')

    def add_category_column(self, category_function: Optional[str], categories_dict: Optional[dict] = None):
        if 'product_name' in self.columns:
            _loc = self.columns.get_loc('product_name') + 1
            if category_function == "first word":
                self.insert(loc=_loc, column='category',
                            value=self.product_name.apply(lambda s: s.split(' ')[0]))
                self.insert(loc=_loc + 1, column='subcategory', value='')
            elif category_function == "regex" and categories_dict is not None:

                def regex_category_function_local(_product_name):
                    return regex_category_function(product_name=_product_name, categories_dict=categories_dict)

                category_df = self[['product_name']].apply(regex_category_function_local,
                                                           axis=1, result_type='expand')
                self.insert(loc=_loc, column='category', value=category_df[0])
                self.insert(loc=_loc + 1, column='subcategory', value=category_df[1])
            elif category_function == "separate_regex" and categories_dict is not None:

                def separate_regex_category_function1_local(_product_name):
                    return separate_regex_category_function1(product_name=_product_name,
                                                             categories_dict=categories_dict)

                def separate_regex_category_function2_local(_product_name):
                    return separate_regex_category_function2(product_name=_product_name,
                                                             categories_dict=categories_dict)

                self.insert(loc=_loc, column='category',
                            value=self['product_name'].apply(separate_regex_category_function1_local))
                self.insert(loc=_loc + 1, column='subcategory',
                            value=self['product_name'].apply(separate_regex_category_function2_local))
            else:
                self.insert(loc=_loc, column='category', value='')
                self.insert(loc=_loc + 1, column='subcategory', value='')

    def add_google_organic_column(self):
        if 'source' in self.columns and 'medium' in self.columns:
            _loc = self.columns.get_loc('medium') + 1
            if len(self) == 0:
                self['is_google_organic'] = None
            else:
                self.insert(loc=_loc,
                            column='is_google_organic',
                            value=self.apply(lambda _df: is_google_organic(source=_df['source'],
                                                                           medium=_df['medium']),
                                             axis=1)
                            )
        else:
            pass

    def add_new_user_column(self):
        if 'user_type' in self.columns:
            _loc = self.columns.get_loc('user_type') + 1
            if len(self) == 0:
                self['is_new_user'] = None
            else:
                self.insert(loc=_loc,
                            column='is_new_user',
                            value=self.apply(lambda _df: is_new_user(user_type=_df['user_type']),
                                             axis=1)
                            )
        else:
            pass

    def add_shopping_stage_all_column(self):
        if 'shopping_stage' in self.columns:
            _loc = self.columns.get_loc('shopping_stage') + 1
            if len(self) == 0:
                self['shopping_stage_all'] = None
            else:
                self.insert(loc=_loc,
                            column='shopping_stage_all',
                            value=self.apply(lambda _df: is_shopping_stage_all(shopping_stage=_df['shopping_stage']),
                                             axis=1)
                            )
        else:
            pass

    def add_has_site_search_column(self):
        if 'search_used' in self.columns:
            _loc = self.columns.get_loc('search_used') + 1
            if len(self) == 0:
                self['has_site_search'] = None
            else:
                self.insert(loc=_loc,
                            column='has_site_search',
                            value=self.apply(lambda _df: has_site_search(search_used=_df['search_used']),
                                             axis=1)
                            )
        else:
            pass

    def add_has_item_column(self):
        if 'item_quantity' in self.columns:
            if len(self) == 0:
                self['has_item'] = None
            else:
                self.insert(loc=self.columns.get_loc('item_quantity') + 1,
                            column='has_item',
                            value=self.item_quantity.apply(lambda quantity: True if quantity>0 else False)
                            )
        else:
            pass

    def add_landing_page_subdomains(self):
        if 'landing_page' in self.columns:
            _loc = self.columns.get_loc('landing_page') + 1
            self.insert(loc=_loc,
                        column='lp_sub1',
                        value=self.landing_page.apply(get_sub1)
                        )
            self.insert(loc=_loc + 1,
                        column='lp_sub2',
                        value=self.landing_page.apply(get_sub2)
                        )
        else:
            pass

    def filter_go(self, inplace=True):
        """filters for results that are only "Google organic".
        """
        obj = self[(self['source'] == 'google') & (self['medium'] == 'organic')].drop(columns=['source', 'medium'])
        if inplace:
            self._update_inplace(obj)
        else:
            return obj

    def bool_column_to_int(self, column_name):
        if column_name in self.columns:
            self[column_name] = self[column_name].apply(lambda b: 1 if b else 0)
            
    
    def add_join_dimensions(self, dimensions: str | list[str]):
        if isinstance(dimensions, str):
            dimensions = [dimensions]
        self.join_dimensions.extend(dimensions)
        
    def remove_join_dimensions(self, dimensions: str | list[str]):
        if isinstance(dimensions, str):
            dimensions = [dimensions]
        for _d in dimensions:
            self.join_dimensions = remove_list_item(self.join_dimensions, _d)

    def fill_nan_with_zeros(self):
        values = {_col: 0 for _col, _type in self.dtypes.items() if _type == 'float' or _type == 'int'}
        self.fillna(value=values, inplace=True)

        values = {_col: 0 for _col in self.columns if all(pd.isna(_v) for _v in self[_col].to_list())}
        self.fillna(value=values, inplace=True)

    def join_on_dimensions(self, dataframe, how: str = "outer"):
        if not set(self.join_dimensions) == set(dataframe.join_dimensions):
            raise ValueError("Input dataframe must have same join_dimensions")

        if self.error:
            _e = self.error
        else:
            _e = dataframe.error

        if _e == 'empty_response':
            _e = None

        _out = GADataFrame(
            df_input=pd.merge(self, dataframe, how=how, on=self.join_dimensions),
            response_type=self.response_type,
            dimensions=list(set(self.dimensions + dataframe.dimensions)),
            metrics=list(set(self.metrics + dataframe.metrics)),
            from_ga_response=False,
            join_dimensions=self.join_dimensions,
            start_date=min(self.date_range.get('start'), dataframe.date_range.get('start')),
            end_date=max(self.date_range.get('end'), dataframe.date_range.get('end')),
            error=_e
        )

        return _out

    def add_row_id(self):
        self.reset_index(drop=True, inplace=True)
        self.reset_index(drop=False, inplace=True)
        self.rename(columns={'index': 'row_id'}, inplace=True)


class GSCDataFrame(pd.DataFrame):
    """
    Subclass of pandas.DataFrame specifically for use with Google Search Console responses.
    The subclass must be initialized with a GSC response object and will create a dataframe
    from the report.

    In addition to pandas DataFrame functionality, the GSCDataFrame has additional metadata attributes:
        dimensions: list of the GSC dimensions
    """

    _metadata = ["dimensions",
                 "metrics",
                 "response_aggregation",
                 "date_range"]

    def __init__(self, df_input,
                 gsc_dimensions: Optional[List[str]] = None,
                 from_gsc_response: bool = False):
        """

        :param df_input:
        :param gsc_dimensions:
        :param from_gsc_response: The init can either come from some date structure that ordinary pandas DataFrame
        can initialize, or a GSC response dictionary
        """

        if gsc_dimensions is None:
            gsc_dimensions = ['country', 'device', 'page', 'query']

        self.dimensions = gsc_dimensions

        if df_input is None:
            self.metrics = ['clicks', 'impressions', 'ctr', 'position']
            super().__init__(None, columns=self.dimensions + self.metrics)
            if 'date' in self.columns:
                self.rename(columns={'date': 'record_date'}, inplace=True)
            if 'page' in self.columns:
                self.rename(columns={'page': 'landing_page'}, inplace=True)
        else:
            super().__init__(df_input)

        if from_gsc_response is True and df_input is not None:
            self.metrics = []
            for metric in ['clicks', 'impressions', 'ctr', 'position']:
                if metric in list(self.columns):
                    self.metrics.append(metric)

            for _i in range(len(self.dimensions) - 1, -1, -1):
                self.insert(loc=0, column=self.dimensions[_i],
                            value=self['keys'].apply(lambda x: x[_i])
                            )
            self.drop(columns='keys', inplace=True)

            if 'query' in self.columns:
                self['query'] = self['query'].apply(lambda s: s.lower())

            if 'page' in self.columns:
                self['landing_page_parameter'] = self['page'].apply(general_utils.url_extract_parameter)
                self['landing_page'] = self['page'].apply(general_utils.strip_url)
                self['landing_page_nodomain'] = self['page'].apply(general_utils.url_strip_domain)
                self.rename(columns={'page': 'landing_page_full'}, inplace=True)

        if from_gsc_response is False and df_input is not None:
            self.dimensions = list(self.columns)
            self.metrics = []
            for metric in ['clicks', 'impressions', 'ctr', 'position']:
                self.dimensions.remove(metric)
                if metric in list(self.columns):
                    self.metrics.append(metric)

        if 'device' in self.columns:
            self.device = self.device.apply(lambda s: s.lower())

        if 'date' in self.columns:
            self.date = pd.to_datetime(self.date, format="%Y-%m-%d")
            self.date = self.date.apply(lambda date_time: datetime.datetime.date(date_time))
            self.rename(columns={'date': 'record_date'}, inplace=True)

        if 'country' in self.columns:
            self.country = self.country.apply(lambda _s: _s.upper())
            self.rename(columns={'country': 'country_iso_code'}, inplace=True)

    def add_question_column(self):
        if 'query' in self.columns:
            self.insert(loc=self.columns.get_loc('query') + 1,
                        column='is_question',
                        value=self['query'].apply(is_question))
        else:
            pass

    def add_transactional_column(self):
        if 'query' in self.columns:
            self.insert(loc=self.columns.get_loc('query') + 1,
                        column='is_transactional',
                        value=self['query'].apply(is_transactional))
        else:
            pass

    def add_investigation_column(self):
        if 'query' in self.columns:
            self.insert(loc=self.columns.get_loc('query') + 1,
                        column='is_investigation',
                        value=self['query'].apply(is_investigation))
        else:
            pass

    def add_branded_column(self, branded_regex: Union[str, Pattern]):
        if 'query' in self.columns and self.branded_regex is not None:

            def is_branded_function(_query):
                if re.match(branded_regex, _query):
                    return True
                else:
                    return False

            self.insert(loc=self.columns.get_loc('query') + 1,
                        column='is_branded',
                        value=self['query'].apply(is_branded_function))
        else:
            pass

    def add_landing_page_subdomains(self):
        if 'landing_page' in self.columns:
            _loc = self.columns.get_loc('landing_page') + 1
            self.insert(loc=_loc,
                        column='lp_sub1',
                        value=self.landing_page.apply(get_sub1)
                        )
            self.insert(loc=_loc + 1,
                        column='lp_sub2',
                        value=self.landing_page.apply(get_sub2)
                        )
        else:
            pass

    def add_category_column(self, category_function: Optional[str], categories_dict: Optional[dict] = None):
        if 'query' in self.columns:
            _loc = self.columns.get_loc('query') + 1
            if category_function == "first word":
                self.insert(loc=_loc, column='category',
                            value=self.product_name.apply(lambda s: s.split(' ')[0]))
                self.insert(loc=_loc + 1, column='subcategory', value='')

            elif category_function == "regex" and categories_dict is not None:

                def regex_category_function_local(_query):
                    return regex_category_function(product_name=_query, categories_dict=categories_dict)

                category_df = self[['query']].apply(regex_category_function_local,
                                                    axis=1, result_type='expand')
                self.insert(loc=_loc, column='category', value=category_df[0])
                self.insert(loc=_loc + 1, column='subcategory', value=category_df[1])

            elif category_function == "separate_regex" and categories_dict is not None:

                def separate_regex_category_function1_local(_query):
                    return separate_regex_category_function1(product_name=_query, categories_dict=categories_dict)

                def separate_regex_category_function2_local(_query):
                    return separate_regex_category_function2(product_name=_query, categories_dict=categories_dict)

                self.insert(loc=_loc, column='category',
                            value=self['query'].apply(separate_regex_category_function1_local))
                self.insert(loc=_loc + 1, column='subcategory',
                            value=self['query'].apply(separate_regex_category_function2_local))
            else:
                self.insert(loc=_loc, column='category', value='')
                self.insert(loc=_loc + 1, column='subcategory', value='')

    def bin_by_position(self,
                        bins=None,
                        reset_index=True) -> pd.DataFrame:
        if bins is None:
            bins = [0, 5, 10, 20, 40, 80, np.Inf]
        _df = self.copy()
        _df['position'] = pd.cut(_df['position'], bins)
        binned_df = _df.groupby('position').sum()[['clicks', 'impressions']]
        binned_df.insert(loc=0, column='queries', value=_df.groupby('position').size())
        if reset_index:
            return binned_df.reset_index()
        else:
            return binned_df

    def add_row_id(self):
        self.reset_index(drop=True, inplace=True)
        self.reset_index(drop=False, inplace=True)
        self.rename(columns={'index': 'row_id'}, inplace=True)


def is_question(query: str) -> bool:
    if general_utils.RE_QUESTION.match(query):
        return True
    else:
        return False


def is_transactional(query: str) -> bool:
    if general_utils.RE_TRANSACTIONAL.match(query):
        return True
    else:
        return False


def is_investigation(query: str) -> bool:
    if general_utils.RE_INVESTIGATION.match(query):
        return True
    else:
        return False


def is_google_organic(source: str,
                      medium: str) -> bool:
    return (source == 'google') and (medium == 'organic')


def is_new_user(user_type: str) -> bool:
    return user_type == 'New Visitor'


def is_shopping_stage_all(shopping_stage: str) -> bool:
    return shopping_stage == 'ALL_VISITS'


def has_site_search(search_used: str) -> bool:
    return search_used == 'Visits With Site Search'


def regex_brand_function(product_name: str, brands_dict: dict):
    for _brand in brands_dict.keys():
        if re.match(brands_dict[_brand], product_name.lower()):
            return _brand
    return 'other'


def regex_category_function(product_name: str, categories_dict: dict):
    for cat in categories_dict.keys():
        for subcat in categories_dict.get(cat).keys():
            if re.match(categories_dict.get(cat).get(subcat),
                        product_name[0].lower()):
                return cat, subcat
    return 'other', 'other'


def separate_regex_category_function1(product_name: str, categories_dict: dict):
    cat_dict = categories_dict.get('categories')
    for cat in cat_dict.keys():
        if re.match(cat_dict.get(cat), product_name.lower()):
            return cat
    return 'other'


def separate_regex_category_function2(product_name: str, categories_dict: dict):
    subcat_dict = categories_dict.get('subcategories')
    for sub_cat in subcat_dict.keys():
        if re.match(subcat_dict.get(sub_cat), product_name.lower()):
            return sub_cat
    return 'other'


def get_response_type(response: dict | RunReportResponse):
    if isinstance(response, RunReportResponse):
        return 'GA4'

    if _t := response.get('response_type'):
        return _t

    # Check for GA3 response
    try:
        rows = response.get('reports', [])[0]['data']['rows']
        dimensions = response.get('reports')[0].get('columnHeader')['dimensions']
        return 'GA3'
    except (TypeError, IndexError, KeyError, AttributeError):
        pass

    # Check for GSC response
    try:
        if {'clicks', 'impressions', 'ctr', 'position'} < set(response.get('rows', [])[0].keys()):
            return 'GSC'
    except (TypeError, IndexError, KeyError, AttributeError):
        pass

    return None


def response_type_from_file(file_path):
    return file_path.split('/')[-3].strip('_results')


def gsc_bin_by_position(df, bins: List[int] = None):
    if bins is None:
        bins = [0, 10, 20, 50, 100]
    _df = df.copy()
    _df['position'] = pd.cut(_df['position'], bins)
    binned_df = _df.groupby('position').sum()[['clicks', 'impressions']]
    binned_df.insert(loc=0, column='queries', value=_df.groupby('position').size())
    return binned_df.reset_index()


def iso_code_2_to_3(iso_code: str):
    if iso_code == "GB":
        return "GBR"
    elif iso_code == "US":
        return "USA"
    else:
        with open(os.path.join(os.path.dirname(__file__), 'data/country_iso_codes.csv'), mode='r') as infile:
            reader = csv.reader(infile)
            isodict = {rows[1]: rows[2] for rows in reader}
        return isodict.get(iso_code, 'ZZZ')


def get_sub_domain(page_path: str, n: int = 1):
    split = page_path.split('/')
    if len(split) > n:
        return page_path.split('/')[n]
    else:
        return ''


def get_sub1(page_path: str):
    return get_sub_domain(page_path=page_path, n=1)


def get_sub2(page_path: str):
    return get_sub_domain(page_path=page_path, n=2)


def remove_list_item(_list, _item) -> list:
    return [_ for _ in _list if _ != _item]




from setuptools import setup

# python setup.py check
# python setup.py sdist
# python setup.py bdist_wheel --universal
# twine upload dist/*

_desc = """PyGoogalytics allows a user to quickly and simply download Google Analytics and Google Search Console data
in the form of a pandas dataframe."""

setup(
    name='pygoogalytics',
    version='0.6.6',
    description=_desc,
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/Blink-SEO/pygoogalytics',
    author='Joshua Prettyman',
    author_email='joshua@blinkseo.co.uk',
    license='MIT',
    packages=['pygoogalytics', 'pygoogalytics.utils'],
    install_requires=[
        'pandas',
        'google-api-python-client>=2.70.0',
        'google-analytics-data==0.16.1',
        'google-ads==20.0.0',
        'google-api-core==2.11.0'
    ],
    classifiers=[
        "Intended Audience :: Developers",
        'Development Status :: 1 - Planning',
        'License :: OSI Approved :: MIT License',
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
    ],
    include_package_data=True,
    package_data={'': ['data/*.csv']},
)



import re

from google.analytics.data_v1beta.types.analytics_data_api import RunReportResponse

from .general_utils import dict_merge

def parse_ga4_response(response: RunReportResponse):
    dimension_headers = [_.name for _ in response.dimension_headers]
    metric_headers = [_.name for _ in response.metric_headers]
    if response.row_count > 0:
        rows = [
            (
                {dim_k: dim_v for dim_k, dim_v in zip(dimension_headers, [_dv.value for _dv in _r.dimension_values])},
                {met_k: float(met_v) for met_k, met_v in zip(metric_headers, [_dv.value for _dv in _r.metric_values])}
            ) for _r in response.rows
        ]
        rows = [dict_merge(t[0], t[1]) for t in rows]
    else:
        rows = []

    _quota = response.property_quota
    _quota_dict = {
        'tokens_per_day': {
            'consumed': _quota.tokens_per_day.consumed,
            'remaining': _quota.tokens_per_day.remaining
        },
        'tokens_per_hour':{
            'consumed': _quota.tokens_per_hour.consumed,
            'remaining': _quota.tokens_per_hour.remaining
        }
    }

    response = {
        'response_type': 'GA4',
        'dimension_headers': dimension_headers,
        'metric_headers': metric_headers,
        'meta_row_count': response.row_count,
        'currency_code': response.metadata.currency_code,
        'time_zone': response.metadata.time_zone,
        'row_count': len(rows),
        'quota': _quota_dict,
        'rows': rows
    }

    return response


def join_ga4_responses(responses: list[dict]) -> dict:
    _rows = []
    for _r in responses:
        _rows.extend(_r.get('rows', []))
    response = {
        'response_type': 'GA4',
        'dimension_headers': responses[0].get('dimension_headers'),
        'metric_headers': responses[0].get('metric_headers'),
        'meta_row_count': responses[0].get('meta_row_count'),
        'currency_code': responses[0].get('currency_code'),
        'time_zone': responses[0].get('time_zone'),
        'row_count': len(_rows),
        'quota': responses[-1].get('quota'),
        'rows': _rows
    }
    return response

def parse_ga3_response(response = None,
              column_header = None,
              response_rows = None,
              report_index: int = 0):

    if response is not None:
        column_header = response.get('reports', [dict()])[report_index].get('columnHeader')
        response_rows = response.get('reports', [dict()])[report_index].get('data', dict()).get('rows')

    if column_header is None or response_rows is None:
        raise KeyError("Incompatible response")

    dimensions = column_header.get('dimensions', [])
    _metrics_headers = column_header.get('metricHeader', dict()).get('metricHeaderEntries', [dict()])
    metrics = [_.get('name') for _ in _metrics_headers]

    dimensions = [remove_ga_prefix(_) for _ in dimensions]
    metrics = [remove_ga_prefix(_) for _ in metrics]

    _dm = dimensions+metrics
    _rows = [_.get('dimensions') + _.get('metrics', [dict()])[0].get('values') for _ in response_rows]
    rows = [{_k: _v for _k, _v in zip(_dm, _row_values)} for _row_values in _rows]

    response = {
        'response_type': 'GA3',
        'dimension_headers': dimensions,
        'metric_headers': metrics,
        'row_count': len(rows),
        'rows': rows
    }

    return response


def remove_ga_prefix(key: str) -> str:
    return re.sub(r'^ga:', '', key)



import datetime
import re
import operator
from typing import Any

from google.ads.googleads.v15.services.types.google_ads_service import GoogleAdsRow

import pandas as pd



def create_google_ads_sql_query(
        resource: str,
        dimensions: list[str],
        conditions: list | None = None,
        custom_condition_expr: str | None = None,
        limit: int | None = None) -> str:

    q = f"""SELECT {', '.join(dimensions)} \nFROM {resource}"""

    _con_expressions = []

    if not conditions:
        conditions = []
    for _c in conditions:
        _con_expressions.append(tuple_to_sql_cond(_c))
    if custom_condition_expr:
         _con_expressions.append(custom_condition_expr)
    if len(_con_expressions) > 0:
        q += " \nWHERE " + " \nAND ".join(_con_expressions)

    if limit:
        q += f" \nLIMIT {limit}"

    return q



def parse_enum(v):
    if isinstance(v, (str, float, bool)) or type(v) == int:
        return v
    if v is None:
        return v
    try:
        _out = v._name_
    except AttributeError:
        if isinstance(v, int):
            return int(v)
        else:
            _out = str(v)
    return _out


def extract_google_ads_rows(rows: list[GoogleAdsRow], dimensions: list[str]) -> list[dict]:
    return [{_dim: parse_enum(operator.attrgetter(_dim)(r)) for _dim in dimensions} for r in rows]


def parse_date(dt: datetime.date | str) -> str:
    if isinstance(dt, datetime.date):
        return dt.strftime("%Y-%m-%d")
    elif re.match(r'\d{4}-\d{2}-\d{2}', dt):
        return dt
    else:
        raise ValueError("date given in incorrect format")


def make_date_condition(dt: datetime.date | str) -> tuple[str, str, Any]:
    if isinstance(dt, tuple):
        return "segments.date", "BETWEEN", dt
    else:
        return "segments.date", "=", dt




def tuple_to_sql_cond(c: tuple) -> str:
    if len(c) == 2:
        c = (c[0], '=', c[1])

    if c[1] == 'BETWEEN' and isinstance(c[2], (tuple, list)):
        if len(c[2]) != 2:
            raise ValueError(f"BETWEEN condition expects 2 arguments, got {c[2]}")
        return f"{c[0]} {c[1]} {parse_sql_input(c[2][0])} AND {parse_sql_input(c[2][1])}"
    else:
        return f"{c[0]} {c[1]} {parse_sql_input(c[2])}"


def parse_sql_input(v) -> str:
    if isinstance(v, str):
        return '"' + v + '"'
    elif isinstance(v, datetime.date):
        return '"' + dt.strftime("%Y-%m-%d") + '"'
    else:
        return v.__repr__()
    # if _c[1] == 'BETWEEN' and isinstance(_c[2], (tuple, list)):
    #     pass


def tidy_column_names(df: pd.DataFrame) -> pd.DataFrame:
    repl = dict()
    for c in df.columns:
        r = c
        if m := re.match(r'metrics\.(\w+)', c):
            r = m.group(1)
        repl[c] = re.sub(r'\.', '/', r)
    return df.rename(columns=repl)



import pandas as pd
import logging
import re
import datetime
import json
from typing import List, Union, Optional, Any

from googleapiclient.errors import HttpError as GoogleApiHttpError
from google.api_core.exceptions import ResourceExhausted, InvalidArgument, PermissionDenied
import google.analytics.data_v1beta.types as ga_data_types

from . import googlepandas as gpd
from .utils import general_utils
from .utils.ga4_parser import parse_ga4_response, parse_ga3_response, join_ga4_responses
from . import pga_logger


class MissingID(AttributeError):
    """Error raised when Property ID is not specified"""
    def __init__(self, message="Missing ID"):
        self.message = message
        super().__init__(self.message)


class GoogalyticsWrapper:
    """
    The GoogalyticsWrapper requires the following arguments to access data:
    - for GSC data: sc_domain. This is the url-like string you see in the Google Search Console web application
    when selecting the site. It is either a full url (e.g. `https://www.example.com/`) or something like `sc_domain:example.com`
    - for GA3 data: the "view_id" you see in "settings" on the GA web application. This is usually an 8- or 9-digit number, passed as a string
    - for GA4 data: the ga4 property id.
    """
    def __init__(self,
                 gsc_resource,
                 ga3_resource,
                 ga4_resource,
                 sc_domain: str = None,
                 view_id: str = None,
                 ga4_property_id: str = None):

        self.sc_domain: str = sc_domain
        self.view_id: str = view_id
        self.ga4_property_id: str = ga4_property_id

        self._api_test_gsc: dict = dict()
        self._api_test_ga3: dict = dict()
        self._api_test_ga4: dict = dict()

        self.gsc_resource = gsc_resource
        self.ga3_resource = ga3_resource
        self.ga4_resource = ga4_resource

        pga_logger.debug(f"initialising GoogalyticsWrapper object")

    # *****************************************************************
    # *** GAPI_WRAPPER STATS ******************************************

    def __dict__(self) -> dict:
        _dates_test = self.available_dates
        gsc_date_range_str = general_utils.date_range_string(dates=_dates_test.get("GSC"),
                                                     alternate_text="No dates available from GSC")
        ga3_date_range_str = general_utils.date_range_string(dates=_dates_test.get("GA3"),
                                                     alternate_text="No dates available from GA3")

        return {
            "API config": {
                "GSC sc-domain": self.sc_domain,
                "GA3 View ID": self.view_id,
                "GA4 Property ID": self.ga4_property_id
            },
            "API status": {
                "GSC status": self.api_test_gsc.get('status'),
                "GA3 status": self.api_test_ga3.get('status'),
                "GSC error": self.api_test_gsc.get('error'),
                "GA3 error": self.api_test_ga3.get('error')
            },
            "Available datas": {
                "GSC": gsc_date_range_str,
                "GA3": ga3_date_range_str
            }
        }

    def __repr__(self):
        _s = "GoogalyticsWrapper object:\n"
        _s += json.dumps(self.__dict__(), indent=2)
        return _s

    @property
    def api_summary(self) -> dict:
        _dates_test = self.available_dates
        _sc_domain = ""
        if re.match("sc-domain:.+", self.sc_domain):
            _sc_domain = self.sc_domain
        return {"GA3 view id": self.view_id,
                "sc-domain": _sc_domain,
                "GA3 API": self.api_test_ga3.get('status'),
                "GSC API": self.api_test_gsc.get('status'),
                "GA3 dates": len(_dates_test.get("GA3")),
                "GSC dates": len(_dates_test.get("GSC")),
                }

    def _perform_api_test_gsc(self) -> dict:
        """test GSC API"""
        pga_logger.debug(f"{self.__class__.__name__}.api_test() :: testing GSC api")

        _api_error = None
        _dates_list = []

        if not self.sc_domain:
            return dict(
                status="SC Domain not set",
                error="SC Domain not set",
                date_count=0,
                timestamp=datetime.datetime.utcnow()
            )

        try:
            _ = self.get_gsc_response(start_date=datetime.date.today() + datetime.timedelta(days=-7),
                                      raise_http_error=True)
            _api_status = "Success"
            pga_logger.debug(f"{self.__class__.__name__}.api_test() :: GSC api successful")
            _dates_list = self.get_dates(result="GSC")
        except GoogleApiHttpError as http_e:
            _api_status = "HttpError"
            _api_error = http_e.reason.split('See also')[0]
            pga_logger.debug(f"{self.__class__.__name__}.api_test() :: GSC api failed")
        except Exception as http_e:
            _api_status = "Other Error"
            _api_error = repr(http_e)
            pga_logger.debug(f"{self.__class__.__name__}.api_test() :: GSC api failed")
            # The HttpError for GSC contains this unhelpful "See also this answer to a question..."
            # which is just a link to an FAQ with a 404 error

        return dict(
            status=_api_status,
            error=_api_error,
            date_count=len(_dates_list),
            timestamp=datetime.datetime.utcnow()
        )

    @property
    def api_test_gsc(self) -> dict:
        if self._api_test_gsc.get('status') is None:
            self._api_test_gsc = self._perform_api_test_gsc()
        return self._api_test_gsc

    @property
    def api_test_ga3(self) -> dict:
        if self._api_test_ga3.get('status') is None or not general_utils.test_time(self._api_test_ga3.get('timestamp'), 3600):
            self._api_test_ga3 = self._perform_api_test_ga3()
        return self._api_test_ga3

    def _perform_api_test_ga3(self) -> dict:
        """test GA API"""
        pga_logger.debug(f"{self.__class__.__name__}.api_test() :: testing GA api")
        _api_error = None
        if not self.view_id:
            return dict(
                status="No View ID",
                error="GA3 View ID not set",
                date_count=0,
                timestamp=datetime.datetime.utcnow()
            )

        _dates_list = []
        try:
            _date = datetime.date.today() + datetime.timedelta(days=-7)
            _ = self.get_ga3_response(
                start_date=_date,
                end_date=_date,
                dimensions=['date'],
                metrics=['users'],
                raise_http_error=True,
                log_error=False
            )
            _api_status = "Success"
            pga_logger.debug(f"{self.__class__.__name__}.api_test() :: GA api successful")
            _dates_list = self.get_dates(result="GA3")
        except GoogleApiHttpError as http_e:
            _api_status = "HttpError"
            _api_error = repr(http_e)
            pga_logger.debug(f"{self.__class__.__name__}.api_test() :: GA api failed")
        except Exception as http_e:
            _api_status = "Other Error"
            _api_error = repr(http_e)
            pga_logger.debug(f"{self.__class__.__name__}.api_test() :: GA api failed")
        return dict(
            status=_api_status,
            error=_api_error,
            date_count=len(_dates_list),
            timestamp=datetime.datetime.utcnow()
        )

    @property
    def api_test_ga4(self) -> dict:
        if self._api_test_ga4.get('status') is None or not general_utils.test_time(self._api_test_ga4.get('timestamp'), 3600):
            self._api_test_ga4 = self._perform_api_test_ga4()
        return self._api_test_ga4

    def _perform_api_test_ga4(self) -> dict:
        """test GA4 API"""
        pga_logger.debug(f"{self.__class__.__name__}.api_test() :: testing GA4 api")

        if not self.ga4_property_id:
            return dict(
                status="No property id",
                error="Ga4 Property ID not set",
                date_count=0,
                timestamp=datetime.datetime.utcnow()
            )

        _dates_list = []
        _error = None
        try:
            _date = datetime.date.today() + datetime.timedelta(days=-7)
            _response = self.get_ga4_response(
                start_date=_date,
                end_date=_date,
                dimensions=['date'],
                metrics=['totalUsers']
            )
            _error = _response.get('error')
            if _error:
                _api_status = _response.get('error_type')
                _api_error = str(_error)
            else:
                _api_status = "Success"
                pga_logger.debug(f"{self.__class__.__name__}.api_test() :: GA4 api successful")
                _dates_list = self.get_dates(result="GA4")
                _api_error = None
        except GoogleApiHttpError as http_e:
            _api_status = "HttpError"
            _api_error = repr(http_e)
            pga_logger.debug(f"{self.__class__.__name__}.api_test() :: GA4 api failed")
        except Exception as http_e:
            _api_status = "Other Error"
            _api_error = repr(http_e)
            pga_logger.debug(f"{self.__class__.__name__}.api_test() :: GA4 api failed")


        return dict(
            status=_api_status,
            error=_api_error,
            date_count=len(_dates_list),
            timestamp=datetime.datetime.utcnow()
        )

    # *****************************************************************************************

    @property
    def available_dates(self) -> dict:
        gsc_date_list: List[datetime.date] = self.get_dates(result="GSC")
        ga3_date_list: List[datetime.date] = self.get_dates(result="GA3")
        ga4_date_list: List[datetime.date] = self.get_dates(result="GA4")
        return {"GA3": ga3_date_list, "GA4": ga4_date_list, "GSC": gsc_date_list}

    # *** Calls to Google API *****************************************************************

    def get_inspection_response(self,
                                inspection_url: str):
        gsc_request = {
            'siteUrl': self.sc_domain,
            'inspectionUrl': inspection_url,
        }
        gsc_response = self.gsc_resource.urlInspection().index().inspect(body=gsc_request).execute()
        return gsc_response

    def get_gsc_response(self,
                         start_date: Union[str, datetime.date],
                         end_date: Optional[Union[str, datetime.date]] = None,
                         gsc_dimensions: Optional[Union[List[str], str]] = None,
                         row_limit: int = 25000,
                         start_row: int = 0,
                         raise_http_error: bool = False,
                         _print_log: bool = False):

        # dimension "searchAppearance" cannot be used alongside any other dimension

        if end_date is None:
            end_date = start_date

        if isinstance(start_date, str):
            start_date = datetime.datetime.strptime(start_date, "%Y-%m-%d").date()
        if isinstance(end_date, str):
            end_date = datetime.datetime.strptime(end_date, "%Y-%m-%d").date()

        start_date_string = start_date.strftime("%Y-%m-%d")
        end_date_string = end_date.strftime("%Y-%m-%d")

        if gsc_dimensions is None:
            gsc_dimensions = ['country', 'device', 'page', 'query']
        elif isinstance(gsc_dimensions, str):
            gsc_dimensions = [gsc_dimensions]

        gsc_request = {
            'startDate': start_date_string,
            'endDate': end_date_string,
            'dimensions': gsc_dimensions,
            # 'dimensionFilterGroups': [{
            #     'filters': [
            #         {'dimension': 'country', 'expression': 'GBR'}
            #     ]
            # }],
            # 'aggregationType': aggregation_type,
            'rowLimit': row_limit,
            'startRow': start_row
        }

        try:
            gsc_response = self.gsc_resource.searchanalytics().query(siteUrl=self.sc_domain,
                                                                     body=gsc_request).execute()
        except GoogleApiHttpError as http_error:
            if re.match(".*user does not have sufficient permissions", repr(http_error).lower()):
                pga_logger.error(
                    f"{self.__class__.__name__}.get_gsc_response() :: user does not have sufficient permissions")
            if raise_http_error:
                raise http_error
            else:
                if _print_log:
                    print(f"{self.__class__.__name__}.get_gsc_response() :: GoogleApiHttpError")
                    print(http_error)
                return None

        try:
            _rows = gsc_response.get("rows", None)
        except AttributeError:
            _rows = None

        if _rows is None:
            pga_logger.debug(f"{self.__class__.__name__}.get_gsc_response() :: empty gsc response")
            if _print_log:
                print(f"{self.__class__.__name__}.get_gsc_response() :: empty gsc response")
            # raise EmptyResponseError("GSC", start_date=start_date, end_date=end_date)
            return None

        return gsc_response

    def get_ga3_response(self,
                         start_date: datetime.date,
                         end_date: datetime.date,
                         dimensions: list[str] = None,
                         metrics: list[str] = None,
                         ga_filters: dict | None = None,
                         raise_http_error: bool = False,
                         log_error: bool = True,
                         filter_google_organic: bool = False,
                         _print_log: bool = False) -> Optional[dict]:

        ga_dimensions = ['ga:' + _s if not re.match(r'ga:', _s) else _s for _s in dimensions]
        ga_metrics = ['ga:'+_s if not re.match(r'ga:', _s) else _s for _s in metrics]

        r = self._ga3_response_raw(
            start_date=start_date,
            end_date=end_date,
            ga_dimensions=ga_dimensions,
            ga_metrics=ga_metrics,
            ga_filters=ga_filters,
            filter_google_organic=filter_google_organic,
            page_token=None
        )

        data = r.get('reports', [{}])[0].get('data', {}).get('rows', [])
        column_header = r.get('reports', [{}])[0].get('columnHeader')
        next_page_token = r.get('reports', [{}])[0].get('nextPageToken')
        error = r.get('error', None)
        error_type = r.get('error_type', None)

        while next_page_token and not error:
            r = self._ga3_response_raw(
                start_date=start_date,
                end_date=end_date,
                ga_dimensions=ga_dimensionsga_dimensions,
                ga_metrics=ga_metrics,
                ga_filters=ga_filters,
                filter_google_organic=filter_google_organic,
                page_token=next_page_token
            )
            error = r.get('error', None)
            error_type = r.get('error_type', None)
            _d = r.get('reports', [dict()])[0].get('data', dict()).get('rows', [])
            data.extend(_d)
            next_page_token = r.get('reports', [{}])[0].get('nextPageToken')

        # print(f"\ndimensions: {ga_dimensions} \n"
        #       f"metrics: {ga_metrics} \n "
        #       f"data: len={len(data)}, \n"
        #       f"error_type: {error_type} \n"
        #       f"column_header: {column_header} \n")

        if column_header is not None:
            response = parse_ga3_response(column_header=column_header, response_rows=data)
        else:
            response = {
                'dimension_headers': dimensions,
                'metric_headers': metrics,
                'row_count': 0,
                'rows': []
            }

        response['response_type'] = 'GA3'
        response['start_date'] = start_date
        response['end_date'] = end_date

        response['error'] = error
        response['error_type'] = error_type

        return response

    def _ga3_response_raw(self,
                          start_date: datetime.date,
                          end_date: datetime.date,
                          ga_dimensions: list[str],
                          ga_metrics: List[str],
                          ga_filters: dict | None = None,
                          filter_google_organic: bool = False,
                          return_raw_response: bool = False,
                          page_token: str = None,
                          _print_log: bool = False):

        if not self.view_id:
            _r = {
                'error': PermissionError("view_id is not set"),
                'error_type': 'missing_view_id'
            }
            return _r

        start_date_string = start_date.strftime("%Y-%m-%d")
        end_date_string = end_date.strftime("%Y-%m-%d")

        _dfc = []  # dimension filter clauses
        _mfc = []  # metric filter clauses
        _orderby = []

        if ga_filters:
            for filter_dict in ga_filters:
                if not isinstance(filter_dict.get('filters'), list):
                    continue
                if len(filter_dict.get('filters')) != 0:
                    if filter_dict.get('filters')[0].get('dimensionName'):
                        _dfc.append(filter_dict)
                    elif filter_dict.get('filters')[0].get('metricName'):
                        _mfc.append(filter_dict)

        if filter_google_organic is True:
            _dfc.append({"operator": 'OR',
                         "filters": [{"dimensionName": 'ga:sourceMedium',
                                      "not": 'false',
                                      "operator": 'EXACT',
                                      "expressions": ['google / organic'],
                                      "caseSensitive": 'false'
                                      }]
                         })

        if 'ga:itemRevenue' in ga_metrics:
            _orderby.append({
                "fieldName": 'ga:itemRevenue',
                "orderType": 'VALUE',
                "sortOrder": 'DESCENDING'
            })

        _request_dict = {
            'viewId': self.view_id,
            'dateRanges': [{'startDate': start_date_string, 'endDate': end_date_string}],
            # 'dimensions': [{'name': 'ga:productName'}],
            # 'metrics': [{'expression': 'ga:itemRevenue'}]
            'dimensions': [{'name': _d} for _d in ga_dimensions],
            "dimensionFilterClauses": _dfc,
            "metricFilterClauses": _mfc,
            "orderBys": _orderby,
            'metrics': [{'expression': _m} for _m in ga_metrics],
            'pageSize': 100_000
        }

        if page_token:
            _request_dict.update({'pageToken': page_token})

        ga3_request = {'reportRequests': [_request_dict]}

        _error = None
        _error_type = None
        try:
            ga3_response = self.ga3_resource.reports().batchGet(body=ga3_request).execute()
            if return_raw_response:
                pga_logger.info(f"{self.__class__.__name__}.get_ga3_response() :: returning raw response")
                return ga3_response
        except GoogleApiHttpError as http_error:
            _error = http_error
            _msg = ''
            if re.match(".*user does not have sufficient permissions", repr(http_error).lower()):
                _error_type = 'insufficient_permissions'
                _msg = f"{self.__class__.__name__}.get_ga3_response() :: user does not have sufficient permissions"
            if re.match(".*viewid must be set", repr(http_error).lower()):
                _error_type = 'missing_view_id'
                _msg = f"{self.__class__.__name__}.get_ga3_response() :: view id is not set"
            ga3_response = None
        except Exception as _e:
            _error = _e
            _error_type = 'other'
            ga3_response = None

        if ga3_response is None:
            ga3_response = dict()
        else:
            try:
                _rows = ga3_response.get('reports', [])[0].get('data').get('rows', None)
            except (AttributeError, KeyError) as _e:
                _rows = None

            if _rows is None:
                _error = AttributeError('ga3_response in incorrect format.')
                _error_type = 'empty_response'
                pga_logger.debug(f"{self.__class__.__name__}.get_ga3_response() :: empty ga response")
                # raise EmptyResponseError("GA3", start_date=start_date, end_date=end_date)

        ga3_response['error'] = _error
        ga3_response['error_type'] = _error_type

        return ga3_response

    def _ga4_response_raw(self,
                         start_date: datetime.date,
                         end_date: datetime.date,
                         ga4_dimensions: list[ga_data_types.Dimension],
                         ga4_metrics: list[ga_data_types.Metric],
                         limit: int,
                         offset: int):

        if not self.ga4_property_id:
            raise MissingID("ga4_property_id is not set")

        request = ga_data_types.RunReportRequest(
            property=f"properties/{self.ga4_property_id}",
            dimensions=ga4_dimensions,
            metrics=ga4_metrics,
            date_ranges=[
                ga_data_types.DateRange(
                    start_date=start_date.strftime("%Y-%m-%d"),
                    end_date=end_date.strftime("%Y-%m-%d")
                )
            ],
            limit=limit,
            offset=offset,
            return_property_quota=True
        )
        ga4_response = self.ga4_resource.run_report(request)

        return ga4_response

    def get_ga4_response(self,
                         start_date: datetime.date,
                         end_date: datetime.date,
                         dimensions: list[str],
                         metrics: list[str],
                         limit: int | None = None) -> (list, dict, Any):

        ga_dimensions = [ga_data_types.Dimension(name=_) for _ in dimensions]
        ga_metrics = [ga_data_types.Metric(name=_) for _ in metrics]

        request_limit = 100_000
        if limit is None:
            limit = 1_000_000_000
        elif limit < 100_000:
            request_limit = limit

        complete: bool = False
        error_type: str | None = None
        responses: list[dict] = []
        error = None
        offset: int = 0

        while not complete:
            num_tries = 0
            success = False
            response = dict()
            while num_tries < 3 and not success:
                error = None
                try:
                    ga4_response = self._ga4_response_raw(
                        start_date=start_date,
                        end_date=end_date,
                        ga4_dimensions=ga_dimensions,
                        ga4_metrics=ga_metrics,
                        limit=request_limit,
                        offset=offset
                    )
                    response = parse_ga4_response(ga4_response)
                    success = True
                except PermissionDenied as _permission_denied_error:
                    complete = True
                    error_type = 'PermissionDenied'
                    error = _permission_denied_error
                    num_tries = 1_000
                except ResourceExhausted as _resource_exhausted_error:
                    error_type = 'ResourceExhausted'
                    complete = True
                    error = _resource_exhausted_error
                    num_tries = 1_000
                except MissingID as _id_error:
                    complete = True
                    error_type = 'MissingID'
                    error = _id_error
                    num_tries = 1_000
                except InvalidArgument as _invalid_argument_error:
                    error_type = 'InvalidArgument'
                    if re.search(r"metrics are incompatible", _invalid_argument_error.message):
                        error_type = 'IncompatibleMetrics'
                    if re.search(r"Invalid property ID", _invalid_argument_error.message):
                        error_type = 'InvalidPropertyID'
                    complete = True
                    error = _invalid_argument_error
                    num_tries = 1_000
                except Exception as _e:
                    error_type = "Error"
                    error = _e
                    num_tries += 1

            if success:
                offset += response.get('row_count', 0)
                if offset >= limit:
                    complete = True
                if response.get('row_count', 0) == 0 or offset >= response.get('meta_row_count', 1_000_000_000):
                    complete = True

                responses.append(response)

        if len(responses) > 1:
            response = join_ga4_responses(responses)
            if response.get('row_count', 0) == 0:
                error = AttributeError("Empty response")
                error_type = "EmptyResponse"
        elif len(responses) == 1:
            response = responses[0]
            if response.get('row_count', 0) == 0:
                error = AttributeError("Empty response")
                error_type = "EmptyResponse"
        else:
            response = {
                'response_type': 'GA4',
                'dimension_headers': dimensions,
                'metric_headers': metrics,
                'rows': []
            }

        response['start_date'] = start_date
        response['end_date'] = end_date

        response['error'] = error
        response['error_type'] = error_type

        return response

    def get_dates(self,
                  result: str,
                  start_date: Optional[Union[datetime.date, str, int]] = None,
                  end_date: Optional[Union[datetime.date, str]] = None,
                  reverse: bool = False) -> List[datetime.date]:

        # set the end_date to yesterday by default.
        # GA data is "available" for today but it is not the whole day.
        if end_date is None:
            end_date = datetime.date.today() + datetime.timedelta(days=-1)

        if isinstance(start_date, str):
            start_date = datetime.datetime.strptime(start_date, '%Y-%m-%d').date()
        if isinstance(end_date, str):
            end_date = datetime.datetime.strptime(end_date, '%Y-%m-%d').date()

        if isinstance(start_date, int):
            start_date = end_date + datetime.timedelta(days=-1 * start_date)

        if re.match(r"GA3", result):
            if start_date is None:
                start_date = datetime.date.today() + datetime.timedelta(days=-1500)
            dimensions = ['ga:date']
            metrics = ['ga:sessions']
        elif re.match(r"GA4", result):
            if start_date is None:
                start_date = datetime.date.today() + datetime.timedelta(days=-1500)
            dimensions = ['date']
            metrics = ['sessions']
        elif re.match(r"GSC", result):
            if start_date is None:
                start_date = datetime.date.today() + datetime.timedelta(days=-500)
            dimensions = ['date']
            metrics = None
        else:
            raise KeyError(f"invalid result {result}")

        _df = self.get_df(result=result,
                          start_date=start_date,
                          end_date=end_date,
                          dimensions=dimensions,
                          metrics=metrics,
                          add_boolean_metrics=False)

        if "record_date" not in _df.columns or len(_df) == 0:
            return []

        if "record_date" not in _df.columns or len(_df) == 0:
            return []
        else:
            return sorted(list(_df["record_date"]), reverse=reverse)

    # *****************************************************************************************
    # *** Return dataframe *****************************************************************

    def get_df(self,
               result: str,
               start_date: Union[str, datetime.date] = None,
               end_date: Optional[Union[str, datetime.date]] = None,
               dimensions: Optional[Union[str, List[str]]] = None,
               metrics: Optional[Union[str, List[str]]] = None,
               row_limit: Optional[int] = None,
               url_list: Optional[Union[str, List[str]]] = None,
               filter_google_organic: bool = False,
               filters: List[dict] = None,
               add_boolean_metrics: bool = False,
               _return_response: bool = False,
               raise_errors: bool = True
               ) -> Union[gpd.GADataFrame, gpd.GSCDataFrame, pd.DataFrame]:
        """
        The `get_df` method accepts the following values for the `result` argument:
        - "GSC": for Google Search Console data
        - "GA3": for Google Analytics 3 (UA) data
        - "URL": for Google Search Console URL inspection data
        - "GA4": for Google Analytics 4 data (note, this is not yet available in production)
        """

        if start_date is None:
            if result == "GSC":
                start_date = datetime.date.today() + datetime.timedelta(days=-3)
            else:
                start_date = datetime.date.today()

        if end_date is None:
            end_date = start_date

        if isinstance(start_date, str):
            start_date = datetime.datetime.strptime(start_date, '%Y-%m-%d').date()
        if isinstance(end_date, str):
            end_date = datetime.datetime.strptime(end_date, '%Y-%m-%d').date()

        if isinstance(dimensions, str):
            dimensions = [dimensions]
        if isinstance(metrics, str):
            metrics = [metrics]
        if isinstance(url_list, str):
            url_list = [url_list]

        if re.match(r"GA4", result):
            return self._get_analytics_df(
                response_type='GA4',
                start_date=start_date,
                end_date=end_date,
                dimensions=dimensions,
                metrics=metrics,
                add_boolean_metrics=add_boolean_metrics,
                limit=row_limit,
                filters=filters,
                return_response=_return_response,
            )
        elif re.match(r"GA3", result):
            return self._get_analytics_df(
                response_type='GA3',
                start_date=start_date,
                end_date=end_date,
                dimensions=dimensions,
                metrics=metrics,
                filters=filters,
                add_boolean_metrics=add_boolean_metrics,
                return_response=_return_response
            )
        elif re.match(r"GSC", result) and result != "GSCQ":
            if row_limit is None:
                row_limit = 100000
            return self._get_gsc_df(start_date=start_date,
                                    end_date=end_date,
                                    gsc_dimensions=dimensions,
                                    row_limit=row_limit,
                                    add_boolean_metrics=add_boolean_metrics)
        elif result == 'GSCQ':
            if row_limit is None:
                row_limit = 100000
            return self._get_gsc_df(start_date=start_date,
                                    end_date=end_date,
                                    gsc_dimensions=['query'],
                                    row_limit=row_limit,
                                    add_boolean_metrics=add_boolean_metrics)
        elif result == 'URL':
            return self._get_urlinspection_df(url_list=url_list)
        else:
            raise KeyError(f"invalid result {result}")

    def _get_gsc_df_raw(self,
                        start_date: datetime.date,
                        end_date: datetime.date,
                        row_limit: int,
                        gsc_dimensions: List[str]) -> Optional[gpd.GSCDataFrame]:

        gsc_response = self.get_gsc_response(start_date=start_date, end_date=end_date,
                                             gsc_dimensions=gsc_dimensions,
                                             row_limit=min(row_limit, 25000))

        if gsc_response is None:
            # Make an empty GSCDataFrame
            gsc_df = gpd.GSCDataFrame(df_input=None,
                                      gsc_dimensions=gsc_dimensions)
            _response_aggregation = None
        else:
            # Make a dataframe of the gsc response
            gsc_df = gpd.from_response(response=gsc_response,
                                       response_type="GSC",
                                       gsc_dimensions=gsc_dimensions)
            _response_aggregation = gsc_df.response_aggregation

        if (row_limit > 25000) and (len(gsc_df) == 25000):
            temp_row_limit = row_limit - 25000
            temp_start_row = 25000  # not 25001: "Zero-based index of the first row in the response"
            frames = [gsc_df]
            while temp_row_limit > 0:

                gsc_response2 = self.get_gsc_response(start_date=start_date, end_date=end_date,
                                                      gsc_dimensions=gsc_dimensions,
                                                      row_limit=min(temp_row_limit, 25000),
                                                      start_row=temp_start_row)
                if gsc_response2 is None:
                    break  # exit the while loop if we get an empty response

                # Make a dataframe of the second gsc response
                new_gsc_df = gpd.from_response(response=gsc_response2,
                                               response_type="GSC",
                                               gsc_dimensions=gsc_dimensions)
                frames.append(new_gsc_df)

                temp_row_limit -= 25000
                temp_start_row += 25000

            if len(frames) > 1:
                gsc_df = gpd.GSCDataFrame(pd.concat(frames, ignore_index=True),
                                          gsc_dimensions=gsc_dimensions)

        gsc_df.response_aggregation = _response_aggregation

        return gsc_df

    def _get_gsc_df(self,
                    start_date: datetime.date,
                    end_date: datetime.date,
                    row_limit: int = 200000,
                    gsc_dimensions: Optional[List[str]] = None,
                    add_boolean_metrics: bool = True) -> Optional[gpd.GSCDataFrame]:

        if gsc_dimensions is None:
            gsc_dimensions = ['date', 'country', 'device', 'page', 'query']

        gsc_df = self._get_gsc_df_raw(start_date=start_date,
                                      end_date=end_date,
                                      row_limit=row_limit,
                                      gsc_dimensions=gsc_dimensions)

        if gsc_df is None:
            return None

        if add_boolean_metrics:
            gsc_df.add_question_column()
            gsc_df.add_transactional_column()
            gsc_df.add_investigation_column()

        return gsc_df

    def _get_analytics_df(self,
                    response_type: str,
                    start_date: datetime.date | str,
                    end_date: datetime.date | str | None = None,
                    dimensions: Optional[List[str]] = None,
                    metrics: list[str] | list[list[str]] = None,
                    add_boolean_metrics: bool = True,
                    filters: Optional[dict] = None,
                    limit: int | None = 100_000_000,
                    return_response: bool = False,
                    raise_errors: bool = False) -> gpd.GADataFrame:

        if dimensions is None:
            dimensions = ['dateHour']
        elif isinstance(dimensions, str):
            dimensions = dimensions.split(';')
        dimensions = [re.sub(r"^ga:", "", _) for _ in dimensions]

        if metrics is None:
            metrics = ['itemRevenue']

        if not end_date:
            end_date = start_date
        if isinstance(start_date, str):
            start_date = general_utils.parse_date(start_date)
        if isinstance(end_date, str):
            end_date = general_utils.parse_date(end_date)
        if end_date < start_date:
            raise ValueError("date range incompatible: end_date < start_date")
        if start_date > datetime.date.today():
            raise ValueError("date range incompatible: start_date in the future")

        if isinstance(metrics, str):
            metrics = metrics.split(';')
        if all(isinstance(_, str) for _ in metrics):
            metrics = [metrics]

        metrics_list: list[list[str]] = []
        for _list in metrics:
            metrics_list.extend([_list[10 * i:10 * i + 10] for i in range((len(_list) - 1) // 10 + 1)])
        metrics_list = [[re.sub(r"^ga:", "", _m) for _m in _sub_list] for _sub_list in metrics_list]

        responses: list = []
        breaking_error: bool = False
        breaking_error_type: str | None = None
        for _metrics in metrics_list:
            if response_type == 'GA3':
                _r = self.get_ga3_response(
                    start_date=start_date,
                    end_date=end_date,
                    dimensions=dimensions,
                    metrics=_metrics,
                    ga_filters=filters,
                    raise_http_error=False
                )
            elif response_type == 'GA4':
                _r = self.get_ga4_response(
                    start_date=start_date,
                    end_date=end_date,
                    dimensions=dimensions,
                    metrics=_metrics,
                    limit=limit
                )
            else:
                raise KeyError("response_type not recognised")

            responses.append(_r)
            if _r.get('error_type') is not None:
                if _r.get('error_type') != 'empty_response':
                    breaking_error = True
                    breaking_error_type = _r.get('error_type')
                    break

        if return_response:
            return responses

        if raise_errors:
            _errors = [_r.get('error') for _r in responses if _r.get('error') is not None]
            if len(_errors) > 0:
                raise _errors[0]

        if breaking_error:
            frames = []
        else:
            frames = [gpd.from_response(response=_r) for _r in responses]

        # if len(frames)>0:
        #     return frames

        if len(frames) == 0:
            dataframe = gpd.GADataFrame(df_input=None,
                                     dimensions=dimensions,
                                     metrics=general_utils.expand_list(metrics_list),
                                     start_date=start_date,
                                     end_date=end_date,
                                     error = breaking_error_type)

        elif all(len(_frame) == 0 for _frame in frames):
            dataframe = gpd.GADataFrame(df_input=None,
                                     dimensions=dimensions,
                                     metrics=general_utils.expand_list(metrics_list),
                                     start_date=start_date,
                                     end_date=end_date,
                                     error = breaking_error_type)
        elif len(frames) == 1:
            dataframe = frames[0]
        else:
            dataframe = frames[0]
            for i in range(1, len(frames)):
                dataframe = dataframe.join_on_dimensions(frames[i], how="outer")
            if len(dataframe) == 0 and not dataframe.error:
                dataframe.error = 'empty_response'

        if add_boolean_metrics:
            dataframe.add_google_organic_column()
            dataframe.add_has_item_column()
            dataframe.add_new_user_column()
            dataframe.add_shopping_stage_all_column()
            dataframe.add_has_site_search_column()

        dataframe.fill_nan_with_zeros()

        return dataframe


    def urlinspection_dict(self, url: str, inspection_index: int = None) -> dict:
        _now = datetime.datetime.utcnow()
        _d = {"record_date": _now.date(),
              "record_time": _now.time(),
              "url": url}

        try:
            response = self.get_inspection_response(url)
        except Exception as _e:
            _d.update({"response": repr(_e)})
            return _d

        _inspectionResult = response.get("inspectionResult")
        if _inspectionResult is None:
            _d.update({"response": "empty"})
            return _d

        _d.update({"response": "success"})

        _indexStatusResult = _inspectionResult.get("indexStatusResult")
        if _indexStatusResult is not None:
            _last_crawl_time = _indexStatusResult.get("lastCrawlTime")
            if _last_crawl_time is not None:
                _last_crawl_time = datetime.datetime.strptime(_last_crawl_time, "%Y-%m-%dT%H:%M:%SZ")
            _d.update({"index_status_result_verdict": _indexStatusResult.get("verdict"),
                       "coverage_state": _indexStatusResult.get("coverageState"),
                       "robotstxt_state": _indexStatusResult.get("robotsTxtState"),
                       "indexing_state": _indexStatusResult.get("indexingState"),
                       "last_crawl_time": _last_crawl_time,
                       "page_fetch_state": _indexStatusResult.get("pageFetchState"),
                       "google_canonical": _indexStatusResult.get("googleCanonical"),
                       "user_canonical": _indexStatusResult.get("userCanonical"),
                       "sitemap": _indexStatusResult.get("sitemap"),
                       "referring_urls": _indexStatusResult.get("referringUrls"),
                       "crawled_as": _indexStatusResult.get("crawledAs")})

        _mobileUsabilityResult = _inspectionResult.get("mobileUsabilityResult")
        if _mobileUsabilityResult is not None:
            _d.update({"mobile_usability_result_verdict": _mobileUsabilityResult.get("verdict")
                       })

        _mobileUsabilityIssue = _inspectionResult.get("mobileUsabilityIssue")
        if _mobileUsabilityIssue is not None:
            _d.update({"mobile_usability_result_verdict": _mobileUsabilityResult.get("verdict")
                       })
        return _d

    def _get_urlinspection_df(self,
                              url_list: List[str]) -> pd.DataFrame:
        if isinstance(url_list, str):
            url_list = [url_list]
        pga_logger.info(f"{self.__class__.__name__}.get_urlinspection_df() :: "
                        f"requesting url inspection for {len(url_list)} urls")
        _frames = []
        for _i, url in enumerate(url_list):
            _frames.append(self.urlinspection_dict(url, inspection_index=_i))

        if len(_frames) == 0:
            return pd.DataFrame()

        df = pd.DataFrame(_frames)

        df.rename(columns={'url': 'url_full'}, inplace=True)

        df['url_parameter'] = df['url_full'].apply(general_utils.url_extract_parameter)
        df['url'] = df['url_full'].apply(general_utils.strip_url)
        df['url_nodomain'] = df['url_full'].apply(general_utils.url_strip_domain)

        return df
